# -*- coding: utf-8 -*-
"""Create, store and reset container mountpoints"""

import os
import stat

from udocker.msg import Msg
from udocker.utils.fileutil import FileUtil


class MountPoint(object):
    """Create, store and reset container mountpoints"""

    orig_dir = "/.mountpoints"

    def __init__(self, localrepo, container_id):
        self.localrepo = localrepo               # LocalRepository object
        self.container_id = container_id         # Container id
        self.mountpoints = {}
        self.container_dir = \
            os.path.realpath(self.localrepo.cd_container(container_id))
        self.container_root = self.container_dir + "/ROOT"
        self.mountpoints_orig_dir = self.container_dir + self.orig_dir
        self.setup()

    def setup(self):
        """Prepare container for mountpoints"""
        if not os.path.isdir(self.mountpoints_orig_dir):
            if not FileUtil(self.mountpoints_orig_dir).mkdir():
                Msg().err("Error: creating dir:", self.mountpoints_orig_dir)
                return False
        return True

    def add(self, cont_path):
        """Add a container relative pathname as destination mountpoint"""
        mountpoint = self.container_root + '/' + cont_path
        orig_mpath = FileUtil(mountpoint).getvalid_path()
        if orig_mpath:
            self.mountpoints[cont_path] = \
                orig_mpath.replace(self.container_root, "", 1)
        if not self.mountpoints[cont_path]:
            self.mountpoints[cont_path] = "/"

    def delete(self, cont_path):
        """Delete container mountpoint"""
        if cont_path not in self.mountpoints or not self.container_root:
            return False
        container_root = os.path.realpath(self.container_root)
        mountpoint = container_root + '/' + cont_path
        orig_mpath = container_root + '/' + self.mountpoints[cont_path]
        mountpoint = os.path.normpath(mountpoint)
        orig_mpath = os.path.normpath(orig_mpath)
        while mountpoint != orig_mpath:
            if mountpoint.startswith(container_root):
                if not FileUtil(mountpoint).remove():
                    Msg().err("Error: while deleting:", cont_path)
                    return False
                link_path = self.mountpoints[cont_path].replace('/', '#')
                FileUtil(self.mountpoints_orig_dir + '/' + link_path).remove()
                mountpoint = os.path.dirname(mountpoint)
        del self.mountpoints[cont_path]
        return True

    def delete_all(self):
        """Delete all mountpoints"""
        for cont_path in dict(self.mountpoints):
            self.delete(cont_path)

    def create(self, host_path, cont_path):
        """Create mountpoint"""
        status = True
        mountpoint = self.container_root + '/' + cont_path
        if os.path.exists(mountpoint):
            if (stat.S_IFMT(os.stat(mountpoint).st_mode) ==
                    stat.S_IFMT(os.stat(host_path).st_mode)):
                return True
            Msg().err("Error: host and container volume paths not same type:",
                      host_path, cont_path)
            return False
        self.add(cont_path)
        if os.path.isfile(host_path):
            FileUtil(os.path.dirname(mountpoint)).mkdir()
            FileUtil(mountpoint).putdata("", "w")
            status = os.path.isfile(mountpoint) or os.path.islink(mountpoint)
        elif os.path.isdir(host_path):
            status = FileUtil(mountpoint).mkdir()
        if not status:
            Msg().err("Error: creating container mountpoint:", cont_path)
            self.delete(cont_path)
            return False
        return True

    def save(self, cont_path):
        """Save one mountpoint"""
        if cont_path not in self.mountpoints:
            return True
        orig_mountpoint = self.mountpoints[cont_path].replace('/', '#')
        curr_mountpoint = cont_path.replace('/', '#')
        curr_mountpoint = self.mountpoints_orig_dir + '/' + curr_mountpoint
        try:
            if not os.path.exists(curr_mountpoint):
                os.symlink(orig_mountpoint, curr_mountpoint)
        except (IOError, OSError):
            return False
        return True

    def save_all(self):
        """Save all mountpoints"""
        for cont_path in self.mountpoints:
            self.save(cont_path)

    def load_all(self):
        """Load all mountpoints"""
        for f_name in os.listdir(self.mountpoints_orig_dir):
            orig_mpath = os.readlink(self.mountpoints_orig_dir + '/' + f_name)
            orig_mpath = orig_mpath.replace('#', '/')
            cont_path = f_name.replace('#', '/')
            self.mountpoints[cont_path] = orig_mpath

    def restore(self):
        """Restore container reverting mountpoints"""
        self.save_all()
        self.load_all()
        self.delete_all()
        FileUtil(self.mountpoints_orig_dir).remove(recursive=True)



# -*- coding: utf-8 -*-
"""runc execution engine"""

import sys
import os
import subprocess
import platform
import stat
import select
import json

from udocker.engine.base import ExecutionEngineCommon
from udocker.engine.proot import PRootEngine
from udocker.msg import Msg
from udocker.config import Config
from udocker.utils.fileutil import FileUtil
from udocker.utils.uvolume import Uvolume
from udocker.engine.nvidia import NvidiaMode
from udocker.utils.filebind import FileBind
from udocker.helper.unique import Unique
from udocker.helper.hostinfo import HostInfo


class RuncEngine(ExecutionEngineCommon):
    """Docker container execution engine using runc
    Provides a namespaces based user space container.
    Inherits from ContainerEngine class
    """

    def __init__(self, localrepo, exec_mode):
        super(RuncEngine, self).__init__(localrepo, exec_mode)
        self.executable = None                   # runc
        self._container_specjson = None
        self._container_specfile = None
        self._container_specdir = self.container_dir
        self._filebind = None
        self.execution_id = None
        self.engine_type = ""

    def select_runc(self):
        """Set runc executable and related variables"""
        self.executable = Config.conf['use_runc_executable']
        if self.executable != "UDOCKER" and not self.executable:
            self.executable = FileUtil("runc").find_exec()

        if self.executable != "UDOCKER" and not self.executable:
            self.executable = FileUtil("crun").find_exec()

        if self.executable == "UDOCKER" or not self.executable:
            self.executable = ""
            arch = HostInfo().arch()
            image_list = []
            eng = ["runc", "crun"]
            if "cgroup2" in FileUtil("/proc/filesystems").getdata('r'):
                eng = ["crun", "runc"]

            if arch == "amd64":
                image_list = [eng[0]+"-x86_64", eng[0],
                              eng[1]+"-x86_64", eng[1]]
            elif arch == "i386":
                image_list = [eng[0]+"-x86", eng[0], eng[1]+"-x86", eng[1]]
            elif arch == "arm64":
                image_list = [eng[0]+"-arm64", eng[0], eng[1]+"-arm64", eng[1]]
            elif arch == "arm":
                image_list = [eng[0]+"-arm", eng[0], eng[1]+"-arm", eng[1]]

            f_util = FileUtil(self.localrepo.bindir)
            self.executable = f_util.find_file_in_dir(image_list)

        if not os.path.exists(self.executable):
            Msg().err("Error: runc/crun executable not found")
            sys.exit(1)
        if "crun" in os.path.basename(self.executable):
            self.engine_type = "crun"
        elif "runc" in os.path.basename(self.executable):
            self.engine_type = "runc"

    def _load_spec(self, new=False):
        """Generate runc spec file"""
        if FileUtil(self._container_specfile).size() != -1 and new:
            FileUtil(self._container_specfile).register_prefix()
            FileUtil(self._container_specfile).remove()

        if FileUtil(self._container_specfile).size() == -1:
            cmd_l = [self.executable, "spec", "--rootless", ]
            status = subprocess.call(cmd_l, shell=False, stderr=Msg.chlderr,
                                     close_fds=True,
                                     cwd=os.path.realpath(\
                                         self._container_specdir))
            if status:
                return False

        json_obj = None
        infile = None
        try:
            infile = open(self._container_specfile, 'r', encoding='utf-8')
            json_obj = json.load(infile)
        except (IOError, OSError, AttributeError, ValueError, TypeError):
            json_obj = None

        if infile:
            infile.close()

        self._container_specjson = json_obj
        return json_obj

    def _save_spec(self):
        """Save spec file"""
        outfile = None
        try:
            outfile = open(self._container_specfile, 'w', encoding='utf-8')
            json.dump(self._container_specjson, outfile)
        except (IOError, OSError, AttributeError, ValueError, TypeError):
            if outfile:
                outfile.close()
            return False
        outfile.close()
        return True

    def _set_spec(self):
        """Set spec values"""
        json_obj = self._container_specjson
        json_obj["root"]["path"] = os.path.realpath(self.container_root)
        json_obj["root"]["readonly"] = False
        if '.' in self.opt["hostname"]:
            json_obj["hostname"] = self.opt["hostname"]
        else:
            json_obj["hostname"] = platform.node()

        if self.opt["cwd"]:
            json_obj["process"]["cwd"] = self.opt["cwd"]

        json_obj["process"]["terminal"] = sys.stdout.isatty()
        if self.engine_type == "crun":
            json_obj["process"]["terminal"] = False

        json_obj["process"]["env"] = []
        for (env_key, env_val) in self.opt["env"]:
            json_obj["process"]["env"].append(f"{env_key}={env_val}")

        json_obj["process"]["args"] = self.opt["cmd"]
        return json_obj

    def _set_id_mappings(self):
        """set uid and gid mappings"""
        json_obj = self._container_specjson
        if "uidMappings" in json_obj["linux"]:
            for idmap in json_obj["linux"]["uidMappings"]:
                if "hostID" in idmap:
                    idmap["hostID"] = HostInfo.uid
        else:
            json_obj["linux"]["uidMappings"] = [ \
                    {"containerID": 0, "hostID": HostInfo.uid, "size":1, }, ]
        if "gidMappings" in json_obj["linux"]:
            for idmap in json_obj["linux"]["gidMappings"]:
                if "hostID" in idmap:
                    idmap["hostID"] = HostInfo.gid
        else:
            json_obj["linux"]["gidMappings"] = [ \
                    {"containerID": 0, "hostID": HostInfo.gid, "size":1, }, ]

    def _del_namespace_spec(self, namespace):
        """Remove a namespace"""
        try:
            json_obj = self._container_specjson
            json_obj["linux"]["namespaces"].remove({"type": namespace})
        except ValueError:
            pass

    def _uid_check(self):
        """Check the uid_map string for container run command"""
        if ("user" in self.opt and self.opt["user"] != "0" and
                self.opt["user"] != "root"):
            Msg().out("Warning: this engine only supports execution as root",
                      l=Msg.WAR)

    def _add_capabilities_spec(self):
        """Set the spec capabilities"""
        if not Config.conf['runc_capabilities']:
            return
        self._container_specjson["process"]["capabilities"]["ambient"] = \
            Config.conf['runc_capabilities']
        self._container_specjson["process"]["capabilities"]["bounding"] = \
            Config.conf['runc_capabilities']
        self._container_specjson["process"]["capabilities"]["effective"] = \
            Config.conf['runc_capabilities']
        self._container_specjson["process"]["capabilities"]["inheritable"] = \
            Config.conf['runc_capabilities']
        self._container_specjson["process"]["capabilities"]["permitted"] = \
            Config.conf['runc_capabilities']

    def _add_device_spec(self, dev_path, mode="rwm"):
        """Add device to the configuration"""
        if not (os.path.exists(dev_path) and dev_path.startswith("/dev/")):
            Msg().err("Error: device not found", dev_path)
            return False
        dev_stat = os.stat(dev_path)
        if stat.S_ISBLK(dev_stat.st_mode):
            dev_type = 'b'
        elif stat.S_ISCHR(dev_stat.st_mode):
            dev_type = 'c'
        else:
            Msg().err("Error: not a device", dev_path)
            return False
        filemode = 0
        if 'r' in mode.lower():
            filemode += 0o444
        if 'w' in mode.lower():
            filemode += 0o222
        if not filemode:
            filemode = 0o666
        if "devices" not in self._container_specjson["linux"]:
            self._container_specjson["linux"]["devices"] = []
        device = {
            "path": dev_path,
            "type": dev_type,
            "major": os.major(dev_stat.st_dev),
            "minor": os.minor(dev_stat.st_dev),
            "fileMode": filemode,
            "uid": HostInfo.uid,
            "gid": HostInfo.gid,
        }
        self._container_specjson["linux"]["devices"].append(device)
        return True

    def _add_devices(self):
        """Add devices to container"""
        added_devices = []
        for dev in self.opt["devices"]:
            dev_name = dev.split(':')[0]
            if self._add_device_spec(dev_name):
                added_devices.append(dev_name)
        if "/dev/nvidiactl" not in added_devices:
            nvidia_mode = NvidiaMode(self.localrepo, self.container_id)
            if nvidia_mode.get_mode():
                for dev_name in nvidia_mode.get_devices():
                    if dev_name not in added_devices:
                        self._add_device_spec(dev_name)

    def _add_mount_spec(self, host_source, cont_dest, rwmode=False,
                        fstype="none", options=None):
        """Add one mount point"""
        if rwmode:
            mode = "rw"
        else:
            mode = "ro"
        mount = {"destination": cont_dest,
                 "type": fstype,
                 "source": host_source,
                 "options": ["rbind", "nosuid", "nodev", mode, ], }
        if options is not None:
            mount["options"] = options
        self._container_specjson["mounts"].append(mount)

    def _del_mount_spec(self, host_source, cont_dest):
        """Remove one mount point"""
        index = self._sel_mount_spec(host_source, cont_dest)
        if index:
            del self._container_specjson["mounts"][index]

    def _sel_mount_spec(self, host_source, cont_dest):
        """Select mount point"""
        for (index, mount) in enumerate(self._container_specjson["mounts"]):
            if (mount["destination"] == cont_dest and
                    mount["source"] == host_source):
                return index
        return None

    def _mod_mount_spec(self, host_source, cont_dest, new):
        """Modify mount spec"""
        index = self._sel_mount_spec(host_source, cont_dest)
        if index is None:
            return False
        mount = self._container_specjson["mounts"][index]
        for new_item in new:
            if new_item == "options":
                if "options" not in mount:
                    mount["options"] = []
                for (new_index, dummy) in enumerate(new["options"]):
                    option_prefix = new["options"][new_index].split('=', 1)[0]
                    for (index, opt) in enumerate(mount["options"]):
                        if opt.startswith(option_prefix):
                            del mount["options"][index]
                    mount["options"] += new["options"]
            else:
                mount[new_item] = new[new_item]
        return True

    def _add_volume_bindings(self):
        """Get the volume bindings string for runc"""
        (host_dir, cont_dir) = self._filebind.start(Config.conf['sysdirs_list'])
        self._add_mount_spec(host_dir, cont_dir, rwmode=True)
        for vol in self.opt["vol"]:
            (host_dir, cont_dir) = Uvolume(vol).split()
            if os.path.isdir(host_dir):
                if host_dir == "/dev":
                    Msg().out("Warning: engine does not support -v",
                              host_dir, l=Msg.WAR)
                    continue
                self._add_mount_spec(host_dir, cont_dir, rwmode=True)
            elif os.path.isfile(host_dir):
                if (host_dir not in Config.conf['sysdirs_list'] and
                        host_dir + ":" + cont_dir not in self.hostauth_list and
                        os.path.dirname(cont_dir) not in
                        Config.conf['mountpoint_prefixes']):
                    Msg().err("Error: engine does not support file mounting:",
                              host_dir)
                else:
                    self._filebind.set_file(host_dir, cont_dir)
                    self._filebind.add_file(host_dir, cont_dir)

    def _run_invalid_options(self):
        """check -p --publish -P --publish-all --net-coop"""
        if self.opt["portsmap"]:
            Msg().out("Warning: this execution mode does not support "
                      "-p --publish", l=Msg.WAR)
        if self.opt["netcoop"]:
            Msg().out("Warning: this execution mode does not support "
                      "-P --netcoop --publish-all", l=Msg.WAR)

    def _proot_overlay(self, proot_mode="P2"):
        """Execute proot within runc"""
        xmode = self.exec_mode.get_mode()
        if xmode not in ("R2", "R3"):
            return False

        preng = PRootEngine(self.localrepo, self.exec_mode)
        preng.exec_mode.force_mode = proot_mode
        preng.select_proot()
        if preng.proot_noseccomp or os.getenv("PROOT_NO_SECCOMP"):
            env_noseccomp = "PROOT_NO_SECCOMP=1"

        if xmode == "R2":
            env_noseccomp = ""

        if xmode == "R3":
            env_noseccomp = "PROOT_NO_SECCOMP=1"

        if env_noseccomp:
            self._container_specjson["process"]["env"].append(env_noseccomp)

        host_executable = preng.executable
        cont_executable = "/.udocker/bin/" + os.path.basename(host_executable)
        self._create_mountpoint(host_executable, cont_executable,
                                dirs_only=False)
        self._filebind.set_file(host_executable, cont_executable)
        self._filebind.add_file(host_executable, cont_executable)
        mode = stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR
        FileUtil(self._filebind.get_path(cont_executable)).chmod(mode)
        self._container_specjson["process"]["args"] = \
                [cont_executable, "-0"] + self.opt["cmd"]
        return True

    def run(self, container_id):
        """Execute a Docker container using runc. This is the main method
        invoked to run the a container with runc.
          * argument: container_id or name
          * options:  many via self.opt see the help
        """

        Config.conf['sysdirs_list'] = (
            "/etc/resolv.conf", "/etc/host.conf",
            # "/etc/passwd", "/etc/group",
        )

        # setup execution
        if not self._run_init(container_id):
            return 2

        self._run_invalid_options()

        self._container_specfile = "config.json"
        if self.container_dir:
            if self.localrepo.iswriteable_container(container_id):
                self._container_specdir = self.container_dir
            else:
                self._container_specdir = FileUtil("SPECDIR").mktmpdir()
                FileUtil(self._container_specdir).register_prefix()
            self._container_specfile = \
                    self._container_specdir + '/' + self._container_specfile

        self._filebind = FileBind(self.localrepo, container_id)
        self._filebind.setup()
        self.select_runc()

        # create new OCI spec file
        if not self._load_spec(new=True):
            return 4

        self._uid_check()

        # if not --hostenv clean the environment
        self._run_env_cleanup_list()

        # set environment variables
        self._run_env_set()
        self._set_spec()
        if (Config.conf['runc_nomqueue'] or
                (Config.conf['runc_nomqueue'] is None and not
                 HostInfo().oskernel_isgreater([4, 8, 0]))):
            self._del_mount_spec("mqueue", "/dev/mqueue")

        self._del_mount_spec("cgroup", "/sys/fs/cgroup")
        self._del_namespace_spec("network")
        self._set_id_mappings()
        self._add_volume_bindings()
        self._add_devices()
        self._add_capabilities_spec()
        self._mod_mount_spec("shm", "/dev/shm", {"options": ["size=2g"]})
        self._proot_overlay()
        self._save_spec()
        if Msg.level >= Msg.DBG:
            runc_debug = ["--debug", ]
            Msg().out(json.dumps(self._container_specjson,
                                 indent=4, sort_keys=True))
        else:
            runc_debug = []

       # build the actual command
        self.execution_id = Unique().uuid(self.container_id)
        cmd_l = self._set_cpu_affinity()
        cmd_l.append(self.executable)
        cmd_l.extend(runc_debug)
        cmd_l.extend(["--root", self._container_specdir, "run"])
        cmd_l.extend(["--bundle", self._container_specdir, self.execution_id])
        Msg().err("CMD =", cmd_l, l=Msg.VER)
        self._run_banner(self.opt["cmd"][0], '%')
        if sys.stdout.isatty():
            return self.run_pty(cmd_l)

        return self.run_nopty(cmd_l)

    def run_pty(self, cmd_l):
        """runc from a terminal"""
        status = subprocess.call(cmd_l, shell=False, close_fds=True)
        self._filebind.finish()
        return status

    def run_nopty(self, cmd_l):
        """runc without a terminal"""
        (pmaster, pslave) = os.openpty()
        status = subprocess.Popen(cmd_l, shell=False, close_fds=True,
                                  stdout=pslave, stderr=pslave)
        os.close(pslave)
        while True:
            status.poll()
            if status.returncode is not None:
                break
            readable, dummy, exception = \
                select.select([pmaster, ], [], [pmaster, ], 5)
            if exception:
                break
            if readable:
                try:
                    if sys.version_info[0] >= 3:
                        sys.stdout.write(os.read(pmaster, 1).decode())
                    else:
                        sys.stdout.write(os.read(pmaster, 1))
                except OSError:
                    break
        try:
            status.terminate()
        except OSError:
            pass
        self._filebind.finish()
        return status



# -*- coding: utf-8 -*-
"""Parser for udocker commands and options"""

import string


class CmdParser(object):
    """Implements a simple command line parser.
    Divides the command into parameters and options
    that can be queried for presence and value.
    The general command format is:
      $ udocker command arg1 arg2 --opta --optb=zzz
    """

    def __init__(self):
        """constructor parses the user command line"""
        self._argv = ""
        self._argv_split = {}
        self._argv_consumed_options = {}
        self._argv_consumed_params = {}
        self._argv_split['CMD'] = ""
        self._argv_split['GEN_OPT'] = []
        self._argv_split['CMD_OPT'] = []
        self._argv_consumed_options['GEN_OPT'] = []
        self._argv_consumed_options['CMD_OPT'] = []
        self._argv_consumed_params['GEN_OPT'] = []
        self._argv_consumed_params['CMD_OPT'] = []

    def parse(self, argv):
        """Parse a command line string.
        Divides the string in three blocks: general_options,
        command name, and command options+arguments
        """
        step = 1
        for arg in argv[1:]:
            if not arg:
                continue

            if step == 1:
                if arg[0] in string.ascii_letters:
                    self._argv_split['CMD'] = arg
                    step = 2
                else:
                    self._argv_split['GEN_OPT'].append(arg)
            elif step == 2:
                self._argv_split['CMD_OPT'].append(arg)

        return step == 2

    def missing_options(self):
        """Get command line options not used/fetched by Cmdp.get()
        """
        all_opt = []
        for pos in range(len(self._argv_split['GEN_OPT'])):
            if (pos not in self._argv_consumed_options['GEN_OPT'] and
                    pos not in self._argv_consumed_params['GEN_OPT']):
                all_opt.append(self._argv_split['GEN_OPT'][pos])

        for pos in range(len(self._argv_split['CMD_OPT'])):
            if (pos not in self._argv_consumed_options['CMD_OPT'] and
                    pos not in self._argv_consumed_params['CMD_OPT']):
                all_opt.append(self._argv_split['CMD_OPT'][pos])

        return all_opt

    def get(self, opt_name, opt_where="CMD_OPT", opt_multiple=False):
        """Get the value of a command line option --xyz=
        multiple=true  multiple occurences of option can be present
        """
        if opt_where == "CMD":
            return self._argv_split["CMD"]

        if opt_where in ("CMD_OPT", "GEN_OPT"):
            if opt_name.startswith('P'):
                return (self._get_param(opt_name,
                                        self._argv_split[opt_where],
                                        self._argv_consumed_options[opt_where],
                                        self._argv_consumed_params[opt_where]))

            if opt_name.startswith('-'):
                return (self._get_option(opt_name,
                                         self._argv_split[opt_where],
                                         self._argv_consumed_options[opt_where],
                                         opt_multiple))

        return None

    def declare_options(self, opts_string, opt_where="CMD_OPT"):
        """Declare in advance options that are part of the command line
        """
        pos = 0
        opt_list = self._argv_split[opt_where]
        while pos < len(opt_list):
            for opt_name in opts_string.strip().split():
                if opt_name.endswith('='):
                    if opt_list[pos].startswith(opt_name):
                        self._argv_consumed_options[opt_where].append(pos)
                    elif opt_list[pos] == opt_name[:-1]:
                        self._argv_consumed_options[opt_where].append(pos)
                        if pos + 1 == len(opt_list):
                            break   # error -x without argument at end of line
                        if (pos < len(opt_list) and
                                not opt_list[pos+1].startswith('-')):
                            self._argv_consumed_options[opt_where].\
                                append(pos + 1)
                elif opt_list[pos] == opt_name:
                    self._argv_consumed_options[opt_where].append(pos)
            pos += 1

    def _get_option(self, opt_name, opt_list, consumed, opt_multiple):
        """Get command line options such as: -x -x= --x --x=
        The options may exist in the first and third part of the udocker
        command line.
        """
        all_args = []
        pos = 0
        list_len = len(opt_list)
        while pos < list_len:
            opt_arg = None
            if ((not opt_list[pos].startswith('-')) and
                    (pos < 1 or (pos not in consumed and not
                                 opt_list[pos-1].endswith('=')))):
                break        # end of options and start of arguments

            if opt_name.endswith('='):
                if opt_list[pos].startswith(opt_name):
                    opt_arg = opt_list[pos].split('=', 1)[1].strip()
                elif (opt_list[pos] == opt_name[:-1] and
                      pos + 1 == list_len):
                    break    # error --arg at end of line
                elif (opt_list[pos] == opt_name[:-1] and
                      not opt_list[pos + 1].startswith('-')):
                    consumed.append(pos)
                    pos += 1
                    opt_arg = opt_list[pos]
            elif opt_list[pos] == opt_name:
                consumed.append(pos)
                opt_arg = True

            pos += 1
            if opt_arg is None:
                continue

            consumed.append(pos-1)
            if opt_multiple:
                all_args.append(opt_arg)
            else:
                return opt_arg

        if opt_multiple:
            return all_args

        return False

    def _get_param(self, opt_name, opt_list, consumed, consumed_params):
        """Get command line parameters
        The CLI of udocker has 3 parts, first options, second command name
        third everything after the command name. The params are the arguments
        that do not start with - and may exist after the options.
        """
        all_args = []
        pos = 0
        param_num = 0
        skip_opts = True
        while pos < len(opt_list):
            if opt_list[pos] == "-":
                skip_opts = False

            if not (skip_opts and
                    (opt_list[pos].startswith('-') or pos in consumed)):
                skip_opts = False
                param_num += 1
                if opt_name[1:] == str(param_num):
                    consumed_params.append(pos)
                    return opt_list[pos]

                if opt_name[1] == '*':
                    consumed_params.append(pos)
                    all_args.append(opt_list[pos])
                elif opt_name[1] == '+' and param_num > 0:
                    consumed_params.append(pos)
                    all_args.append(opt_list[pos])
            pos += 1

        if opt_name[1] == '*':
            return all_args

        if opt_name[1] == '+':
            return all_args[1:]

        return None



#!/usr/bin/env python

# -*- coding: utf-8 -*-
"""
udocker unit tests: GetURL
"""

from unittest import TestCase, main
from unittest.mock import patch
from udocker.utils.curl import GetURL
from udocker.utils.curl import GetURLpyCurl
from udocker.utils.curl import GetURLexeCurl
from udocker.config import Config


class GetURLTestCase(TestCase):
    """Test GetURL() perform http operations portably."""

    def setUp(self):
        Config().getconf()
        Config().conf['timeout'] = 1
        Config().conf['ctimeout'] = 1
        Config().conf['download_timeout'] = 1
        Config().conf['http_agent'] = ""
        Config().conf['http_proxy'] = ""
        Config().conf['http_insecure'] = 0
        Config().conf['use_curl_exec'] = ""

    def tearDown(self):
        pass

    def _get(self, *args, **kwargs):
        """Mock for pycurl.get."""
        return args[0]

    @patch.object(GetURLexeCurl, '_select_implementation')
    @patch.object(GetURLexeCurl, 'is_available')
    @patch.object(GetURLpyCurl, 'is_available')
    @patch('udocker.utils.curl.Msg')
    def test_01_init(self, mock_msg, mock_gupycurl,
                     mock_guexecurl, mock_select):
        """Test01 GetURL() constructor."""
        mock_msg.level = 0
        mock_gupycurl.return_value = False
        mock_guexecurl.return_value = True
        geturl = GetURL()
        mock_select.assert_called()
        self.assertEqual(geturl.ctimeout, Config().conf['ctimeout'])
        self.assertEqual(geturl.insecure, Config().conf['http_insecure'])
        self.assertFalse(geturl.cache_support)

    @patch('udocker.utils.curl.Msg')
    @patch.object(GetURLexeCurl, 'is_available')
    @patch.object(GetURLpyCurl, 'is_available')
    def test_02__select_implementation(self, mock_gupycurl,
                                       mock_guexecurl, mock_msg):
        """Test02 GetURL()._select_implementation()."""
        Config.conf['use_curl_executable'] = ""
        mock_msg.level = 0
        mock_gupycurl.return_value = True
        geturl = GetURL()
        geturl._select_implementation()
        self.assertTrue(geturl.cache_support)
        self.assertTrue(mock_gupycurl.called)

        mock_gupycurl.return_value = False
        geturl = GetURL()
        geturl._select_implementation()
        self.assertFalse(geturl.cache_support)

        mock_gupycurl.return_value = False
        mock_guexecurl.return_value = False
        with self.assertRaises(NameError) as nameerr:
            geturl = GetURL()
            geturl._select_implementation()
            self.assertEqual(nameerr.exception.code, 1)

    def test_03_get_content_length(self):
        """Test03 GetURL().get_content_length()."""
        hdr = type('test', (object,), {})()
        hdr.data = {"content-length": 10, }
        geturl = GetURL()
        self.assertEqual(geturl.get_content_length(hdr), 10)

        hdr = type('test', (object,), {})()
        hdr.data = {"content-length": dict(), }
        geturl = GetURL()
        self.assertEqual(geturl.get_content_length(hdr), -1)

    @patch.object(GetURLpyCurl, 'is_available')
    def test_04_set_insecure(self, mock_gupycurl):
        """Test04 GetURL().set_insecure()."""
        mock_gupycurl.return_value = True
        geturl = GetURL()
        geturl.set_insecure()
        self.assertEqual(geturl.insecure, True)

        geturl = GetURL()
        geturl.set_insecure(False)
        self.assertEqual(geturl.insecure, False)

    @patch.object(GetURLpyCurl, 'is_available')
    def test_05_set_proxy(self, mock_gupycurl):
        """Test05 GetURL().set_proxy()."""
        mock_gupycurl.return_value = True
        geturl = GetURL()
        geturl.set_proxy("http://host")
        self.assertEqual(geturl.http_proxy, "http://host")

    def test_06_get(self):
        """Test06 GetURL().get()."""
        geturl = GetURL()
        self.assertRaises(TypeError, geturl.get)

        geturl = GetURL()
        geturl._geturl = type('test', (object,), {})()
        geturl._geturl.get = self._get
        self.assertEqual(geturl.get("http://host"), "http://host")

    def test_07_post(self):
        """Test07 GetURL().post()."""
        geturl = GetURL()
        self.assertRaises(TypeError, geturl.post)
        self.assertRaises(TypeError, geturl.post, "http://host")

        geturl = GetURL()
        geturl._geturl = type('test', (object,), {})()
        geturl._geturl.get = self._get
        status = geturl.post("http://host", {"DATA": 1, })
        self.assertEqual(status, "http://host")

    def test_08_get_status_code(self):
        """Test08 GetURL().get_status_code()."""
        sline = "HTTP-Version 400 Reason-Phrase"
        geturl = GetURL()
        status = geturl.get_status_code(sline)
        self.assertEqual(status, 400)

        sline = "HTTP-Version Reason-Phrase"
        geturl = GetURL()
        status = geturl.get_status_code(sline)
        self.assertEqual(status, 400)

        sline = ""
        geturl = GetURL()
        status = geturl.get_status_code(sline)
        self.assertEqual(status, 404)


if __name__ == '__main__':
    main()



#!/usr/bin/env python

# -*- coding: utf-8 -*-
"""
udocker unit tests: HostInfo
"""

import pwd
from unittest import TestCase, main
from unittest.mock import patch
from udocker.helper.hostinfo import HostInfo


class HostInfoTestCase(TestCase):
    """Test HostInfo"""

    @patch('udocker.helper.hostinfo.os.getgid')
    @patch('udocker.helper.hostinfo.os.getuid')
    @patch('udocker.helper.hostinfo.pwd.getpwuid')
    def test_01_username(self, mock_getpwuid, mock_uid, mock_gid):
        """Test01 HostInfo().username."""
        usr = pwd.struct_passwd(["root", "*", "0", "0", "root usr",
                                 "/root", "/bin/bash"])
        mock_uid.return_value = 0
        mock_gid.return_value = 0
        mock_getpwuid.return_value = usr
        name = HostInfo().username()
        self.assertEqual(name, usr.pw_name)

    @patch('udocker.helper.hostinfo.platform.architecture')
    @patch('udocker.helper.hostinfo.platform.machine')
    def test_02_arch(self, mock_mach, mock_arch):
        """Test02 HostInfo().arch."""
        mock_mach.return_value = "x86_64"
        mock_arch.return_value = ('64bit', '')
        result = HostInfo().arch()
        self.assertEqual(result, "amd64")

        mock_mach.return_value = "x86_64"
        mock_arch.return_value = ('32bit', '')
        result = HostInfo().arch()
        self.assertEqual(result, "i386")

        mock_mach.return_value = "arm"
        mock_arch.return_value = ('64bit', '')
        result = HostInfo().arch()
        self.assertEqual(result, "arm64")

        mock_mach.return_value = "arm"
        mock_arch.return_value = ('32bit', '')
        result = HostInfo().arch()
        self.assertEqual(result, "arm")

    @patch('udocker.helper.hostinfo.platform.system')
    def test_03_osversion(self, mock_sys):
        """Test03 HostInfo().osversion."""
        mock_sys.return_value = "linux"
        result = HostInfo().osversion()
        self.assertEqual(result, "linux")

    @patch('udocker.helper.hostinfo.platform.release')
    def test_04_oskernel(self, mock_rel):
        """Test04 HostInfo().oskernel."""
        mock_rel.return_value = "3.2.1"
        result = HostInfo().oskernel()
        self.assertEqual(result, "3.2.1")

    @patch.object(HostInfo, 'oskernel')
    def test_05_oskernel_isgreater(self, mock_kernel):
        """Test05 HostInfo().oskernel_isgreater."""
        mock_kernel.return_value = "1.1.2-"
        status = HostInfo().oskernel_isgreater([1, 1, 1])
        self.assertTrue(status)

        mock_kernel.return_value = "1.2.1-"
        status = HostInfo().oskernel_isgreater([1, 1, 1])
        self.assertTrue(status)

        mock_kernel.return_value = "1.0.0-"
        status = HostInfo().oskernel_isgreater([1, 1, 1])
        self.assertFalse(status)

    def test_06_cmd_has_option(self):
        """Test06 HostInfo().cmd_has_option."""
        status = HostInfo().cmd_has_option("ls", "-a")
        self.assertTrue(status)

        status = HostInfo().cmd_has_option("ls", "-z")
        self.assertFalse(status)

    @patch('udocker.helper.hostinfo.Uprocess.check_output')
    def test_07_termsize(self, mock_chkout):
        """Test07 HostInfo().termsize."""
        mock_chkout.return_value = "24 80"
        status = HostInfo().termsize()
        self.assertEqual(status, (24, 80))


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""Information about the host"""

import os
import re
import pwd
import platform

from udocker import is_genstr
from udocker.utils.uprocess import Uprocess

class HostInfo(object):
    """Get information from the host system"""

    uid = os.getuid()
    gid = os.getgid()

    def username(self):
        """Get username"""
        try:
            return pwd.getpwuid(self.uid).pw_name
        except KeyError:
            return ""

    def arch(self):
        """Get the host system architecture"""
        arch = ""
        try:
            machine = platform.machine()
            bits = platform.architecture()[0]
            if machine == "x86_64":
                if bits == "32bit":
                    arch = "i386"
                else:
                    arch = "amd64"
            elif machine in ("i386", "i486", "i586", "i686"):
                arch = "i386"
            elif machine.startswith("arm") or machine.startswith("aarch"):
                if bits == "32bit":
                    arch = "arm"
                else:
                    arch = "arm64"
        except (NameError, AttributeError):
            pass
        return arch

    def osversion(self):
        """Get operating system"""
        try:
            return platform.system().lower()
        except (NameError, AttributeError):
            return ""

    def oskernel(self):
        """Get operating system"""
        try:
            return platform.release()
        except (NameError, AttributeError):
            return "6.1.1"

    def oskernel_isgreater(self, version):
        """Compare kernel version is greater or equal than ref_version
        version: list [int, int, int]
        """
        match = re.search(r"([0-9.]+)", self.oskernel())
        if match:
            os_release = match.group(1)
        else:
            return True

        for (idx, os_version) in enumerate(os_release.split('.')):
            if idx >= len(version):
                break
            if int(os_version) > int(version[idx]):
                return True
            if int(os_version) < int(version[idx]):
                return False

        return True

    def cmd_has_option(self, executable, search_option, arg=None):
        """Check if executable has a given cli option"""
        if not executable:
            return False
        arg_list = []
        if arg and is_genstr(arg):
            arg_list = [arg]
        elif isinstance(arg, list):
            arg_list = arg
        out = Uprocess().get_output([executable] + arg_list + ["--help"])
        if out and search_option in re.split(r"[=|\*\[\]\n,; ]+", out):
            return True
        return False

    def termsize(self):
        """Get guest operating system terminal size"""
        try:
            with open("/dev/tty") as tty:
                cmd = ['stty', 'size']
                lines, cols = Uprocess().check_output(cmd, stdin=tty).split()
                return (int(lines), int(cols))
        except (OSError, IOError):
            pass
        return (24, 80)



#!/usr/bin/env python
"""
udocker unit tests: LocalFileAPI
"""

from unittest import TestCase, main
from unittest.mock import patch, Mock
from udocker.localfile import LocalFileAPI


class LocalFileAPITestCase(TestCase):
    """Test LocalFileAPI()."""

    def setUp(self):
        str_local = 'udocker.container.localrepo.LocalRepository'
        self.lrepo = patch(str_local)
        self.local = self.lrepo.start()
        self.mock_lrepo = Mock()
        self.local.return_value = self.mock_lrepo

    def tearDown(self):
        self.lrepo.stop()

    # def test_01__init(self):
    #     """Test01 LocalFileAPI() constructor."""

    @patch('udocker.localfile.OciLocalFileAPI.load')
    @patch('udocker.localfile.DockerLocalFileAPI.load')
    @patch.object(LocalFileAPI, '_get_imagedir_type')
    @patch('udocker.localfile.FileUtil.remove')
    @patch.object(LocalFileAPI, '_untar_saved_container')
    @patch('udocker.localfile.os.makedirs')
    @patch('udocker.localfile.FileUtil.mktmp')
    @patch('udocker.localfile.Msg.err')
    @patch('udocker.localfile.os.path.exists')
    def test_02_load(self, mock_exists, mock_msg, mock_mktmp, mock_mkdir,
                     mock_untar, mock_remove, mock_imgtype, mock_dockerload,
                     mock_ociload):
        """Test02 LocalFileAPI().load."""
        mock_exists.return_value = False
        status = LocalFileAPI(self.mock_lrepo).load('imgfile')
        self.assertTrue(mock_msg.called)
        self.assertFalse(status)

        mock_exists.return_value = True
        mock_mktmp.return_value = '/tmp/imgdir'
        mock_mkdir.return_value = None
        mock_untar.return_value = False
        mock_remove.return_value = None
        status = LocalFileAPI(self.mock_lrepo).load('imgfile')
        self.assertTrue(mock_mktmp.called)
        self.assertTrue(mock_mkdir.called)
        self.assertTrue(mock_remove.called)
        self.assertTrue(mock_msg.called)
        self.assertFalse(status)

        mock_exists.return_value = True
        mock_mktmp.return_value = '/tmp/imgdir'
        mock_mkdir.return_value = None
        mock_untar.return_value = True
        mock_remove.return_value = None
        mock_imgtype.return_value = 'Docker'
        mock_dockerload.return_value = ['docker-repo1', 'docker-repo2']
        status = LocalFileAPI(self.mock_lrepo).load('imgfile')
        self.assertTrue(mock_imgtype.called)
        self.assertTrue(mock_dockerload.called)
        self.assertEqual(status, ['docker-repo1', 'docker-repo2'])

        mock_exists.return_value = True
        mock_mktmp.return_value = '/tmp/imgdir'
        mock_mkdir.return_value = None
        mock_untar.return_value = True
        mock_remove.return_value = None
        mock_imgtype.return_value = 'OCI'
        mock_ociload.return_value = ['OCI-repo1', 'OCI-repo2']
        status = LocalFileAPI(self.mock_lrepo).load('imgfile')
        self.assertTrue(mock_imgtype.called)
        self.assertTrue(mock_ociload.called)
        self.assertEqual(status, ['OCI-repo1', 'OCI-repo2'])

        mock_exists.return_value = True
        mock_mktmp.return_value = '/tmp/imgdir'
        mock_mkdir.return_value = None
        mock_untar.return_value = True
        mock_remove.return_value = None
        mock_imgtype.return_value = ''
        status = LocalFileAPI(self.mock_lrepo).load('imgfile')
        self.assertTrue(mock_imgtype.called)
        self.assertEqual(status, [])

    @patch('udocker.localfile.DockerLocalFileAPI.save')
    def test_03_save(self, mock_dockersave):
        """Test03 LocalFileAPI().save."""
        mock_dockersave.return_value = True
        img_tag = ['tag1', 'tag2']
        status = LocalFileAPI(self.mock_lrepo).save(img_tag, 'imgfile')
        self.assertTrue(status)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
udocker unit tests: UdockerTools
"""

import tarfile
from tarfile import TarInfo
from unittest import TestCase, main
from unittest.mock import Mock, patch
from io import StringIO
from udocker.config import Config
from udocker.utils.curl import CurlHeader
from udocker.tools import UdockerTools

BUILTINS = "builtins"
BOPEN = BUILTINS + '.open'


class UdockerToolsTestCase(TestCase):
    """Test UdockerTools()."""

    def setUp(self):
        Config().getconf()
        str_local = 'udocker.container.localrepo.LocalRepository'
        self.lrepo = patch(str_local)
        self.local = self.lrepo.start()
        self.mock_lrepo = Mock()
        self.local.return_value = self.mock_lrepo

    def tearDown(self):
        self.lrepo.stop()

    @patch('udocker.tools.GetURL')
    def test_01_init(self, mock_geturl):
        """Test01 UdockerTools() constructor."""
        mock_geturl.return_value = None
        utools = UdockerTools(self.local)
        self.assertTrue(mock_geturl.called)
        self.assertEqual(utools.localrepo, self.local)

    @patch('udocker.tools.Msg')
    def test_02__instructions(self, mock_msg):
        """Test02 UdockerTools()._instructions()."""
        utools = UdockerTools(self.local)
        utools._instructions()
        self.assertTrue(mock_msg.return_value.out.call_count, 2)

    def test_03__version2int(self):
        """Test03 UdockerTools()._version2int()."""
        utools = UdockerTools(self.local)
        status = utools._version2int("2.4")
        self.assertEqual(status, 2004000)

    def test_04__version_isok(self):
        """Test04 UdockerTools()._version_isok()."""
        Config.conf['tarball_release'] = "1.3"
        utools = UdockerTools(self.local)
        status = utools._version_isok("2.4")
        self.assertTrue(status)

        Config.conf['tarball_release'] = "2.3"
        utools = UdockerTools(self.local)
        status = utools._version_isok("1.4")
        self.assertFalse(status)

    @patch('udocker.tools.FileUtil.getdata')
    def test_05_is_available(self, mock_fuget):
        """Test05 UdockerTools().is_available()."""
        Config.conf['tarball_release'] = "2.3"
        mock_fuget.return_value = "2.3\n"
        utools = UdockerTools(self.local)
        status = utools.is_available()
        self.assertTrue(status)

    @patch('udocker.tools.FileUtil.remove')
    @patch('udocker.tools.FileUtil.register_prefix')
    @patch('udocker.tools.os.listdir')
    def test_06_purge(self, mock_lsdir, mock_fureg, mock_furm):
        """Test06 UdockerTools().purge()."""
        mock_lsdir.side_effect = [["f1", "f2"],
                                  ["f3", "f4"],
                                  ["f5", "f6"]]
        mock_fureg.side_effect = [None, None, None, None, None, None]
        mock_furm.side_effect = [None, None, None, None, None, None]
        utools = UdockerTools(self.local)
        utools.purge()
        self.assertTrue(mock_lsdir.call_count, 3)
        self.assertTrue(mock_fureg.call_count, 4)
        self.assertTrue(mock_furm.call_count, 4)

    @patch('udocker.tools.GetURL.get')
    @patch('udocker.tools.FileUtil.remove')
    @patch('udocker.tools.FileUtil.mktmp')
    def test_07__download(self, mock_fumktmp, mock_furm, mock_geturl):
        """Test07 UdockerTools()._download()."""
        url = "https://down/file"
        hdr = CurlHeader()
        hdr.data["X-ND-HTTPSTATUS"] = "HTTP/1.1 200 OK"
        mock_fumktmp.return_value = "/tmp/tmpf"
        mock_furm.return_value = None
        mock_geturl.return_value = (hdr, "content type...")
        utools = UdockerTools(self.local)
        status = utools._download(url)
        self.assertTrue(mock_fumktmp.called)
        self.assertEqual(status, "/tmp/tmpf")

        url = "https://down/file"
        hdr = CurlHeader()
        hdr.data["X-ND-HTTPSTATUS"] = "HTTP/1.1 401 FAIL"
        mock_fumktmp.return_value = "/tmp/tmpf"
        mock_furm.return_value = None
        mock_geturl.return_value = (hdr, "content type...")
        utools = UdockerTools(self.local)
        status = utools._download(url)
        self.assertTrue(mock_furm.called)
        self.assertEqual(status, "")

    @patch('udocker.tools.os.path.isfile')
    @patch('udocker.tools.os.path.realpath')
    @patch('udocker.tools.os.path.exists')
    @patch.object(UdockerTools, '_download')
    def test_08__get_file(self, mock_downl, mock_exists, mock_rpath,
                          mock_isfile):
        """Test08 UdockerTools()._get_file()."""
        url = ""
        mock_downl.return_value = ""
        mock_exists.return_value = False
        mock_isfile.return_value = False
        utools = UdockerTools(self.local)
        status = utools._get_file(url)
        self.assertFalse(mock_downl.called)
        self.assertTrue(mock_exists.called)
        self.assertEqual(status, "")

        url = "https://down/file"
        mock_downl.return_value = "/tmp/file"
        mock_exists.return_value = True
        mock_isfile.return_value = True
        mock_rpath.return_value = "/tmp/file"
        utools = UdockerTools(self.local)
        status = utools._get_file(url)
        self.assertTrue(mock_downl.called)
        self.assertTrue(mock_exists.called)
        self.assertTrue(mock_isfile.called)
        self.assertEqual(status, "/tmp/file")

    @patch.object(UdockerTools, '_version_isok')
    @patch('udocker.tools.FileUtil.remove')
    @patch('udocker.tools.FileUtil.getdata')
    @patch('udocker.tools.os.path.basename')
    @patch('udocker.tools.FileUtil.mktmpdir')
    @patch('udocker.tools.os.path.isfile')
    def test_09__verify_version(self, mock_isfile, mock_fumktmp,
                                mock_osbase, mock_fugetdata,
                                mock_furm, mock_versionok):
        """Test09 UdockerTools()._verify_version()."""
        tball = "/home/udocker.tar"
        mock_isfile.return_value = False
        utools = UdockerTools(self.local)
        status = utools._verify_version(tball)
        self.assertTrue(mock_isfile.called)
        self.assertEqual(status, (False, ""))

        tball = "/home/udocker.tar"
        mock_isfile.return_value = True
        mock_fumktmp.return_value = ""
        utools = UdockerTools(self.local)
        status = utools._verify_version(tball)
        self.assertTrue(mock_isfile.called)
        self.assertTrue(mock_fumktmp.called)
        self.assertEqual(status, (False, ""))

        tball = "/home/udocker.tar"
        tinfo1 = TarInfo("udocker_dir/lib/VERSION")
        tinfo2 = TarInfo("a")
        mock_isfile.return_value = True
        mock_fumktmp.return_value = "/home/tmp"
        mock_osbase.return_value = "VERSION"
        mock_fugetdata.return_value = "1.2.7"
        mock_furm.return_value = None
        mock_versionok.return_value = True
        with patch.object(tarfile, 'open', autospec=True) as open_mock:
            open_mock.return_value.getmembers.return_value = [tinfo2, tinfo1]
            open_mock.return_value.extract.return_value = None
            utools = UdockerTools(self.local)
            status = utools._verify_version(tball)
            self.assertEqual(status, (True, "1.2.7"))
            self.assertTrue(mock_furm.called)

    @patch('udocker.tools.os.path.basename')
    @patch('udocker.tools.FileUtil')
    @patch('udocker.tools.os.path.isfile')
    def test_10__install(self, mock_isfile, mock_futil,
                         mock_osbase):
        """Test10 UdockerTools()._install()."""
        tfile = ""
        mock_isfile.return_value = False
        utools = UdockerTools(self.local)
        status = utools._install(tfile)
        self.assertFalse(status)

        tinfo1 = TarInfo("udocker_dir/bin/ls")
        tinfo2 = TarInfo("udocker_dir/lib/lib1")
        tfile = "udocker.tar"
        mock_isfile.return_value = True
        mock_futil.return_value.chmod.return_value = None
        mock_futil.return_value.rchmod.side_effect = [None, None, None,
                                                      None, None, None]
        mock_osbase.side_effect = ["ls", "ls", "lib1", "lib1", "doc", "doc1"]
        self.local.create_repo.return_value = None
        with patch.object(tarfile, 'open', autospec=True) as open_mock:
            open_mock.return_value.getmembers.side_effect = [[tinfo1, tinfo2],
                                                             [tinfo1, tinfo2],
                                                             [tinfo1, tinfo2]]
            open_mock.return_value.extract.side_effect = [None, None]
            utools = UdockerTools(self.local)
            status = utools._install(tfile)
            self.assertTrue(status)
            self.assertTrue(mock_futil.called)
            self.assertTrue(mock_futil.return_value.rchmod.call_count, 4)

    def test_11__get_mirrors(self):
        """Test11 UdockerTools()._get_mirrors()."""
        mirrors = "https://download.ncg.ingrid.pt/udocker-1.2.7.tar.gz"
        utools = UdockerTools(self.local)
        status = utools._get_mirrors(mirrors)
        self.assertEqual(status, [mirrors])

    @patch.object(UdockerTools, '_get_file')
    @patch.object(UdockerTools, '_get_mirrors')
    @patch('udocker.tools.json.load')
    def test_12_get_installinfo(self, mock_jload, mock_mirr, mock_getf):
        """Test12 UdockerTools().get_installinfo()."""
        Config.conf['installinfo'] = "/home/info.json"
        res = {"tarversion": "1.2.7"}
        mock_jload.return_value = {"tarversion": "1.2.7"}
        mock_mirr.return_value = ["/home/info.json"]
        mock_getf.return_value = "info.json"
        subuid_line = StringIO('{"tarversion": "1.2.7"}')
        with patch(BOPEN) as mopen:
            mopen.return_value.__iter__ = (
                lambda self: iter(subuid_line.readline, ''))
            utools = UdockerTools(self.local)
            status = utools.get_installinfo()
            self.assertEqual(status, res)

    @patch.object(UdockerTools, '_install')
    @patch.object(UdockerTools, '_verify_version')
    @patch.object(UdockerTools, '_get_file')
    @patch.object(UdockerTools, '_get_mirrors')
    @patch('udocker.tools.FileUtil.remove')
    def test_13__install_logic(self, mock_furm, mock_getmirr, mock_getfile,
                               mock_verversion, mock_install):
        """Test13 UdockerTools()._install_logic()."""
        mock_furm.return_value = None
        mock_getmirr.return_value = "https://down.pt/udocker-1.2.7.tar.gz"
        mock_getfile.return_value = "udocker-1.2.7.tar.gz"
        mock_verversion.return_value = (True, "1.2.7")
        mock_install.return_value = True
        utools = UdockerTools(self.local)
        status = utools._install_logic(False)
        self.assertTrue(status)
        self.assertTrue(mock_getmirr.called)
        self.assertTrue(mock_getfile.called)
        self.assertTrue(mock_verversion.called)
        self.assertTrue(mock_install.called)

        mock_furm.return_value = None
        mock_getmirr.return_value = "https://down.pt/udocker-1.2.7.tar.gz"
        mock_getfile.return_value = "udocker-1.2.7.tar.gz"
        mock_verversion.return_value = (False, "")
        mock_install.return_value = True
        utools = UdockerTools(self.local)
        status = utools._install_logic(True)
        self.assertTrue(status)

        mock_furm.return_value = None
        mock_getmirr.return_value = "https://down.pt/udocker-1.2.7.tar.gz"
        mock_getfile.return_value = "udocker-1.2.7.tar.gz"
        mock_verversion.return_value = (False, "")
        mock_install.return_value = True
        utools = UdockerTools(self.local)
        status = utools._install_logic(False)
        self.assertFalse(status)

    @patch('udocker.tools.Msg')
    @patch.object(UdockerTools, 'get_installinfo')
    @patch.object(UdockerTools, '_install_logic')
    @patch.object(UdockerTools, 'is_available')
    def test_14_install(self, mock_isavail, mock_instlog,
                        mock_getinfo, mock_msg):
        """Test14 UdockerTools().install()."""
        mock_msg.level = 0
        Config.conf['autoinstall'] = True
        Config.conf['tarball'] = "udocker-1.2.7.tar.gz"
        Config.conf['tarball_release'] = "1.2.7"
        Config.conf['installretry'] = 2
        mock_isavail.return_value = True
        utools = UdockerTools(self.local)
        status = utools.install(False)
        self.assertTrue(status)
        self.assertTrue(mock_isavail.called)

        Config.conf['autoinstall'] = False
        Config.conf['tarball'] = "udocker-1.2.7.tar.gz"
        Config.conf['tarball_release'] = "1.2.7"
        Config.conf['installretry'] = 2
        mock_isavail.return_value = False
        utools = UdockerTools(self.local)
        status = utools.install(False)
        self.assertEqual(status, None)
        self.assertFalse(utools._autoinstall)

        Config.conf['autoinstall'] = False
        Config.conf['tarball'] = ""
        Config.conf['tarball_release'] = "1.2.7"
        Config.conf['installretry'] = 2
        mock_isavail.return_value = True
        utools = UdockerTools(self.local)
        status = utools.install(False)
        self.assertTrue(status)
        self.assertEqual(utools._tarball, "")

        Config.conf['autoinstall'] = True
        Config.conf['tarball'] = "udocker-1.2.7.tar.gz"
        Config.conf['tarball_release'] = "1.2.7"
        Config.conf['installretry'] = 2
        mock_isavail.return_value = False
        mock_instlog.side_effect = [False, True]
        mock_getinfo.side_effect = [None, None]
        utools = UdockerTools(self.local)
        status = utools.install(False)
        self.assertTrue(status)
        self.assertTrue(mock_instlog.call_count, 2)
        self.assertTrue(mock_getinfo.call_count, 2)

        Config.conf['autoinstall'] = True
        Config.conf['tarball'] = "udocker-1.2.7.tar.gz"
        Config.conf['tarball_release'] = "1.2.7"
        Config.conf['installretry'] = 2
        mock_isavail.return_value = False
        mock_instlog.side_effect = [False, False]
        mock_getinfo.side_effect = [None, None]
        utools = UdockerTools(self.local)
        status = utools.install(False)
        self.assertFalse(status)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
udocker unit tests: DockerLocalFileAPI
"""

from unittest import TestCase, main
from unittest.mock import patch, Mock
from udocker.docker import DockerLocalFileAPI
from udocker.config import Config


class DockerLocalFileAPITestCase(TestCase):
    """Test DockerLocalFileAPI() manipulate Docker images."""

    def setUp(self):
        Config().getconf()
        str_local = 'udocker.container.localrepo.LocalRepository'
        self.lrepo = patch(str_local)
        self.local = self.lrepo.start()
        self.mock_lrepo = Mock()
        self.local.return_value = self.mock_lrepo

    def tearDown(self):
        self.lrepo.stop()

    def test_01_init(self):
        """Test01 DockerLocalFileAPI() constructor."""
        dlocapi = DockerLocalFileAPI(self.local)
        self.assertEqual(dlocapi.localrepo, self.local)

    @patch('udocker.docker.os.listdir')
    @patch('udocker.docker.FileUtil.isdir')
    def test_02__load_structure(self, mock_isdir, mock_ldir):
        """Test02 DockerLocalFileAPI()._load_structure()."""
        res = {'repoconfigs': {}, 'repolayers': {}}
        mock_ldir.return_value = ["xx"]
        dlocapi = DockerLocalFileAPI(self.local)
        structure = dlocapi._load_structure("/tmp")
        self.assertEqual(structure, res)

        mock_isdir.return_value = False
        mock_ldir.return_value = ["repositories", ]
        self.local.load_json.return_value = {"REPO": "", }
        dlocapi = DockerLocalFileAPI(self.local)
        structure = dlocapi._load_structure("/tmp")
        res = {'repolayers': {}, 'repoconfigs': {},
               'repositories': {'REPO': ''}}
        self.assertEqual(structure, res)

        mock_isdir.return_value = False
        mock_ldir.return_value = ["manifest.json", ]
        dlocapi = DockerLocalFileAPI(self.local)
        structure = dlocapi._load_structure("/tmp")
        res = {'repolayers': {}, 'repoconfigs': {},
               'manifest': {'REPO': ''}}
        self.assertEqual(structure, res)

        jfname = "x" * 70 + ".json"
        res = {"repolayers": {},
               "repoconfigs": {jfname: {"json": {"k": "v"},
                                        "json_f": "/tmp/"+jfname}}}
        mock_isdir.return_value = False
        mock_ldir.return_value = [jfname, ]
        self.local.load_json.return_value = {"k": "v"}
        dlocapi = DockerLocalFileAPI(self.local)
        structure = dlocapi._load_structure("/tmp")
        self.assertEqual(structure, res)

        fname = "x" * 64
        res = {"repolayers": {fname: {"VERSION": {"k": "v"}}},
               "repoconfigs": dict()}
        mock_isdir.return_value = True
        mock_ldir.side_effect = [[fname, ], ["VERSION", ], ]
        self.local.load_json.return_value = {"k": "v"}
        dlocapi = DockerLocalFileAPI(self.local)
        structure = dlocapi._load_structure("/tmp")
        self.assertEqual(structure, res)

        fname = "x" * 64
        fulllayer = "/tmp/" + fname + "/json"
        res = {"repolayers": {fname: {"json": {"k": "v"},
                                      "json_f": fulllayer}},
               "repoconfigs": dict()}
        mock_isdir.return_value = True
        mock_ldir.side_effect = [[fname, ], ["json", ], ]
        self.local.load_json.return_value = {"k": "v"}
        dlocapi = DockerLocalFileAPI(self.local)
        structure = dlocapi._load_structure("/tmp")
        self.assertEqual(structure, res)

        fname = "x" * 64
        fulllayer = "/tmp/" + fname + "/layer1"
        res = {"repolayers": {fname: {"layer_f": fulllayer}},
               "repoconfigs": dict()}
        mock_isdir.return_value = True
        mock_ldir.side_effect = [[fname, ], ["layer1", ], ]
        self.local.load_json.return_value = {"k": "v"}
        dlocapi = DockerLocalFileAPI(self.local)
        structure = dlocapi._load_structure("/tmp")
        self.assertEqual(structure, res)

    def test_03__find_top_layer_id(self):
        """Test03 DockerLocalFileAPI()._find_top_layer_id()."""
        fname = "x" * 64
        fulllayer = "/tmp/" + fname + "/layer1"
        structure = dict()
        dlocapi = DockerLocalFileAPI(self.local)
        structure = dlocapi._find_top_layer_id(structure)
        self.assertEqual(structure, "")

        lid = "12345"
        fulllayer = "/tmp/" + lid + "/json"
        struc = {"repolayers": {lid: {"json": {"parent": "v1"},
                                      "json_f": fulllayer}},
                 "repoconfigs": dict()}
        dlocapi = DockerLocalFileAPI(self.local)
        status = dlocapi._find_top_layer_id(struc, "v1")
        self.assertEqual(status, lid)

    def test_04__sorted_layers(self):
        """Test04 DockerLocalFileAPI()._sorted_layers()."""
        lid = "12345"
        fulllayer = "/tmp/" + lid + "/json"
        struc = {"repolayers": {lid: {"json": {"layer": lid},
                                      "json_f": fulllayer}},
                 "repoconfigs": dict()}
        dlocapi = DockerLocalFileAPI(self.local)
        status = dlocapi._sorted_layers(struc, lid)
        self.assertEqual(status, [lid])

    def test_05__get_from_manifest(self):
        """Test05 DockerLocalFileAPI()._get_from_manifest()."""
        struc = dict()
        imgtag = "image:tag"
        dlocapi = DockerLocalFileAPI(self.local)
        status = dlocapi._get_from_manifest(struc, imgtag)
        self.assertEqual(status, ("", []))

        imgtag = "IMAGE"
        struc = {"manifest": [{"RepoTags": {"IMAGE": {"TAG": "tag"}},
                               "Layers": ["l1", "l2"],
                               "Config": "conf"}]}
        dlocapi = DockerLocalFileAPI(self.local)
        status = dlocapi._get_from_manifest(struc, imgtag)
        self.assertEqual(status, ("conf", ["l2", "l1"]))

    @patch('udocker.docker.Msg')
    @patch('udocker.docker.CommonLocalFileApi._move_layer_to_v1repo')
    @patch.object(DockerLocalFileAPI, '_sorted_layers')
    @patch.object(DockerLocalFileAPI, '_find_top_layer_id')
    @patch.object(DockerLocalFileAPI, '_get_from_manifest')
    def test_06__load_image_step2(self, mock_manif, mock_findtop,
                                  mock_sort, mock_mvlayer, mock_msg):
        """Test06 DockerLocalFileAPI()._load_image_step2()."""
        struc = dict()
        imgrepo = "img1"
        tag = "tag1"
        mock_msg.level = 0
        mock_manif.return_value = ("", [])
        mock_findtop.return_value = ""
        mock_sort.return_value = list()
        self.local.save_json.return_value = None
        dlocapi = DockerLocalFileAPI(self.local)
        dlocapi._imagerepo = imgrepo
        status = dlocapi._load_image_step2(struc, imgrepo, tag)
        self.assertEqual(status, ["img1:tag1"])

        lid = "l2"
        fulllayer = "/tmp/" + lid + "/json"
        struc = {"manifest": [{"RepoTags": {"IMAGE": {"TAG": "tag"}},
                               "Layers": ["l2"],
                               "Config": "conf"}],
                 "repolayers": {lid: {"json": {"layer": lid},
                                      "json_f": fulllayer,
                                      "VERSION": "2.0"}},
                 "repoconfigs": {"conf.json": {"json": {"layer": lid},
                                               "json_f": fulllayer}}}
        imgrepo = "img1"
        tag = "tag1"
        mock_manif.return_value = ("conf.json", ["l2"])
        mock_findtop.return_value = "l2"
        mock_sort.return_value = ["l2"]
        mock_mvlayer.side_effect = [True, True]
        self.local.save_json.return_value = None
        dlocapi = DockerLocalFileAPI(self.local)
        dlocapi._imagerepo = imgrepo
        status = dlocapi._load_image_step2(struc, imgrepo, tag)
        self.assertEqual(status, [])

    @patch('udocker.docker.CommonLocalFileApi._load_image')
    def test_07__load_repositories(self, mock_loadi):
        """Test07 DockerLocalFileAPI()._load_repositories()."""
        struct = dict()
        dlocapi = DockerLocalFileAPI(self.local)
        status = dlocapi._load_repositories(struct)
        self.assertFalse(status)

        struct = {"repositories": dict()}
        dlocapi = DockerLocalFileAPI(self.local)
        status = dlocapi._load_repositories(struct)
        self.assertEqual(status, list())

        structure = {"repositories": {"IMAGE": {"TAG": "tag", }, }, }
        mock_loadi.return_value = ["image:tag"]
        dlocapi = DockerLocalFileAPI(self.local)
        status = dlocapi._load_repositories(structure)
        self.assertEqual(status, ["image:tag"])

    @patch('udocker.docker.Msg')
    @patch.object(DockerLocalFileAPI, '_load_repositories')
    @patch.object(DockerLocalFileAPI, '_load_structure')
    @patch('udocker.docker.CommonLocalFileApi._load_image')
    @patch('udocker.docker.Unique')
    def test_08_load(self, mock_unique, mock_loadimg,
                     mock_lstruct, mock_lrepo, mock_msg):
        """Test08 DockerLocalFileAPI().load()."""
        tmp_imgdir = "/tmp/img1"
        mock_msg.level = 0
        mock_lstruct.return_value = dict()
        dlocapi = DockerLocalFileAPI(self.local)
        status = dlocapi.load(tmp_imgdir)
        self.assertEqual(status, [])

        tmp_imgdir = "/tmp/img1"
        fname = "x" * 64
        fulllayer = "/tmp/" + fname + "/layer1"
        struc = {"repolayers": {fname: {"layer_f": fulllayer}},
                 "repositories": ["repo1", "repo2"]}
        mock_lstruct.return_value = struc
        mock_lrepo.return_value = ["repo1", "repo2"]
        dlocapi = DockerLocalFileAPI(self.local)
        status = dlocapi.load(tmp_imgdir)
        self.assertEqual(status, ["repo1", "repo2"])

        struc = {"repolayers": {fname: {"layer_f": fulllayer}},
                 "manifest": dict()}
        mock_lstruct.return_value = struc
        mock_unique.return_value.imagename.return_value = "xyz123"
        mock_unique.return_value.imagetag.return_value = "tag1"
        mock_loadimg.return_value = ["repo1", "repo2"]
        dlocapi = DockerLocalFileAPI(self.local)
        status = dlocapi.load(tmp_imgdir)
        self.assertEqual(status, ["repo1", "repo2"])

    @patch('udocker.docker.FileUtil.putdata')
    @patch('udocker.docker.CommonLocalFileApi.create_container_meta')
    @patch('udocker.docker.FileUtil.copyto')
    @patch('udocker.docker.FileUtil.mkdir')
    @patch('udocker.docker.os.path.exists')
    @patch('udocker.docker.os.path.basename')
    @patch('udocker.docker.FileUtil.rename')
    @patch('udocker.docker.ChkSUM.sha256')
    def test_09__save_image(self, mock_sha256, mock_rename, mock_osbase,
                            mock_exists, mock_mkdir, mock_copyto,
                            mock_meta, mock_put):
        """Test09 DockerLocalFileAPI()._save_image()."""
        imgrepo = "img1"
        tag = "tag1"
        struct = dict()
        tmp_imgdir = "/tmp/img1"
        manifest = dict()
        self.local.cd_imagerepo.return_value = None
        self.local.get_image_attributes.return_value = \
            (manifest, ['/cont/lname1.layer'])
        self.local.save_json.side_effect = [False, True]
        dlocapi = DockerLocalFileAPI(self.local)
        status = dlocapi._save_image(imgrepo, tag, struct, tmp_imgdir)
        self.assertFalse(status)

        imgrepo = "IMAGE"
        tag = "tag1"
        fname = "x" * 64
        fulllayer = "/tmp/" + fname + "/layer1"
        struc = {"repolayers": {fname: {"layer_f": fulllayer}},
                 "repositories": {"IMAGE": {"TAG": "tag1", }, }}
        tmp_imgdir = "/tmp/img1"
        manifest = {
            "fsLayers": ({"blobSum": "foolayername"},),
            "history": ({"v1Compatibility": '["foojsonstring"]'},)
        }
        mock_sha256.return_value = \
            "8a29a15cefaeccf6545f7ecf11298f9672d2f0cdaf9e357a95133ac3ad3e1f07"
        mock_rename.return_value = None
        mock_osbase.return_value = "lname1.layer"
        mock_exists.return_value = False
        mock_mkdir.return_value = None
        mock_copyto.return_value = False
        self.local.cd_imagerepo.return_value = None
        self.local.get_image_attributes.return_value = \
            (manifest, ['/cont/lname1.layer'])
        self.local.save_json.side_effect = [True, True]
        dlocapi = DockerLocalFileAPI(self.local)
        status = dlocapi._save_image(imgrepo, tag, struc, tmp_imgdir)
        self.assertFalse(status)

        imgrepo = "repo1"
        tag = "tag1"
        fname = "x" * 64
        fulllayer = "/tmp/" + fname + "/lname1.layer"
        manifest = {
            "fsLayers": ({"blobSum": "foolayername"},),
            "history": ({"v1Compatibility": '["foojsonstring"]'},)
        }
        struc = {"repolayers": {fname: {"layer_f": fulllayer}},
                 "repositories": {"repo1": {"TAG": "tag1", }, },
                 "manifest": [manifest]}
        tmp_imgdir = "/tmp/img1"
        mock_sha256.return_value = \
            "8a29a15cefaeccf6545f7ecf11298f9672d2f0cdaf9e357a95133ac3ad3e1f07"
        mock_rename.return_value = None
        mock_osbase.return_value = "lname1.layer"
        mock_exists.return_value = False
        mock_mkdir.return_value = None
        mock_copyto.return_value = True
        mock_meta.return_value = dict()
        self.local.save_json.return_value = True
        mock_put.return_value = True
        self.local.cd_imagerepo.return_value = None
        self.local.get_image_attributes.return_value = \
            (manifest, ['/cont/lname1.layer'])
        self.local.save_json.side_effect = [True, True]
        dlocapi = DockerLocalFileAPI(self.local)
        status = dlocapi._save_image(imgrepo, tag, struc, tmp_imgdir)
        self.assertTrue(status)

    @patch('udocker.docker.Msg')
    @patch.object(DockerLocalFileAPI, '_save_image')
    @patch('udocker.docker.FileUtil.remove')
    @patch('udocker.docker.FileUtil.tar')
    @patch('udocker.docker.os.makedirs')
    @patch('udocker.docker.FileUtil.mktmp')
    def test_10_save(self, mock_mktmp, mock_mkdir, mock_tar,
                     mock_rm, mock_svimg, mock_msg):
        """Test10 DockerLocalFileAPI().save()."""
        imglist = list()
        imgfile = ""
        mock_msg.level = 0
        mock_mktmp.return_value = "/tmp/img1"
        mock_mkdir.return_value = None
        mock_rm.return_value = None
        self.local.save_json.side_effect = [True, True]
        dlocapi = DockerLocalFileAPI(self.local)
        status = dlocapi.save(imglist, imgfile)
        self.assertFalse(status)

        imglist = [("/img1", "tag1"),]
        imgfile = ""
        mock_mktmp.return_value = "/tmp/img1"
        mock_mkdir.return_value = None
        mock_rm.return_value = None
        mock_svimg.return_value = True
        mock_tar.return_value = False
        self.local.save_json.side_effect = [True, True]
        dlocapi = DockerLocalFileAPI(self.local)
        status = dlocapi.save(imglist, imgfile)
        self.assertFalse(status)

        imglist = [("/img1", "tag1"),]
        imgfile = ""
        mock_mktmp.return_value = "/tmp/img1"
        mock_mkdir.return_value = None
        mock_rm.return_value = None
        mock_svimg.return_value = True
        mock_tar.return_value = True
        self.local.save_json.side_effect = [True, True]
        dlocapi = DockerLocalFileAPI(self.local)
        status = dlocapi.save(imglist, imgfile)
        self.assertTrue(status)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
=============
udocker setup
=============
Wrapper to execute basic docker containers without using docker.
This tool is a last resort for the execution of docker containers
where docker is unavailable. It only provides a limited set of
functionalities.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from setuptools import setup, find_packages
from udocker import __version__

with open('README.md', 'r', encoding='utf8') as readme_file:
    README = readme_file.read()

with open('CHANGELOG.md', 'r', encoding='utf8') as history_file:
    HISTORY = history_file.read()

CONF_DIR = '/udocker/etc'
REQUIREMENTS = []
SETUP_REQUIREMENTS = ['pytest-runner', ]
TEST_REQUIREMENTS = ['pytest', ]

setup(
    author="Jorge Gomes",
    author_email='udocker@lip.pt',
    version=__version__,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: System',
        'Topic :: Utilities',
    ],
    description="A basic user tool to execute simple docker \
        containers in batch or interactive systems without root privileges",
    entry_points={
        'console_scripts': ['udocker=udocker.maincmd:main'],
    },
    install_requires=REQUIREMENTS,
    license="Apache Software License 2.0",
    long_description=README + '\n\n' + HISTORY,
    long_description_content_type='text/markdown',
    include_package_data=True,
    keywords='Linux containers, HPC on cloud, Virtualization',
    name='udocker',
    packages=find_packages(),
    data_files=[(CONF_DIR, ['etc/udocker.conf'])],
    setup_requires=SETUP_REQUIREMENTS,
    test_suite='tests',
    tests_require=TEST_REQUIREMENTS,
    url='https://github.com/indigo-dc/udocker',
    zip_safe=False,
)



#!/usr/bin/env python
"""
udocker unit tests: FileBind
"""

from unittest import TestCase, main
from unittest.mock import Mock, patch
from udocker.utils.filebind import FileBind
from udocker.config import Config


class FileBindTestCase(TestCase):
    """Test FileBind()."""

    def setUp(self):
        Config().getconf()
        self.bind_dir = "/.bind_host_files"
        self.orig_dir = "/.bind_orig_files"
        str_local = 'udocker.container.localrepo.LocalRepository'
        self.lrepo = patch(str_local)
        self.local = self.lrepo.start()
        self.mock_lrepo = Mock()
        self.local.return_value = self.mock_lrepo

    def tearDown(self):
        self.lrepo.stop()

    @patch('udocker.utils.filebind.os.path.realpath')
    def test_01_init(self, mock_realpath):
        """Test01 FileBind() constructor"""
        container_id = "CONTAINERID"
        mock_realpath.return_value = "/tmp"
        fbind = FileBind(self.local, container_id)
        self.assertEqual(fbind.localrepo, self.local)
        self.assertEqual(fbind.container_id, container_id)
        self.assertTrue(mock_realpath.called)
        self.assertTrue(fbind.container_root, fbind.container_dir + "/ROOT")
        self.assertTrue(
            fbind.container_bind_dir, fbind.container_root + self.bind_dir)
        self.assertTrue(
            fbind.container_orig_dir, fbind.container_dir + self.orig_dir)
        self.assertIsNone(fbind.host_bind_dir)

    @patch('udocker.utils.filebind.Msg')
    @patch('udocker.utils.filebind.os.path.isdir')
    @patch('udocker.utils.filebind.os.path.realpath')
    @patch('udocker.utils.filebind.FileUtil')
    def test_02_setup(self, mock_futil, mock_realpath, mock_isdir,
                      mock_msg):
        """Test02 FileBind().setup()."""
        mock_msg.level = 0
        container_id = "CONTAINERID"
        mock_realpath.return_value = "/tmp"
        mock_isdir.side_effect = [True, True]
        status = FileBind(self.local, container_id).setup()
        self.assertTrue(mock_isdir.called)
        self.assertTrue(status)

        mock_isdir.side_effect = [False, True]
        mock_futil.return_value.mkdir.return_value = False
        status = FileBind(self.local, container_id).setup()
        self.assertFalse(status)

        mock_isdir.side_effect = [True, False]
        mock_futil.return_value.mkdir.return_value = False
        status = FileBind(self.local, container_id).setup()
        self.assertFalse(status)

    @patch('udocker.utils.filebind.os.path.isdir')
    @patch('udocker.utils.filebind.os.path.islink')
    @patch('udocker.utils.filebind.os.path.isfile')
    @patch('udocker.utils.filebind.os.path.realpath')
    @patch('udocker.utils.filebind.os.listdir')
    @patch('udocker.utils.filebind.FileUtil')
    def test_03_restore(self, mock_futil, mock_listdir,
                        mock_realpath, mock_isfile,
                        mock_islink, mock_isdir):
        """Test03 FileBind().restore()."""
        container_id = "CONTAINERID"
        mock_realpath.return_value = "/tmp"
        mock_listdir.return_value = []
        mock_futil.return_value.remove.return_value = True

        mock_isdir.return_value = False
        fbind = FileBind(self.local, container_id)
        fbind.restore()
        self.assertFalse(mock_listdir.called)

        mock_isdir.return_value = True
        fbind = FileBind(self.local, container_id)
        fbind.restore()
        self.assertTrue(mock_listdir.called)

        mock_listdir.return_value = ["is_file1", "is_dir", "is_file2"]
        mock_isfile.side_effect = [True, False, True]
        mock_islink.side_effect = [True, False, False]
        fbind = FileBind(self.local, container_id)
        fbind.restore()
        self.assertTrue(mock_isfile.called)
        self.assertTrue(mock_islink.called)

    @patch.object(FileBind, 'setup')
    @patch('udocker.utils.filebind.FileUtil.mktmpdir')
    @patch('udocker.utils.filebind.os.path.realpath')
    @patch('udocker.utils.filebind.os.path.isfile')
    @patch('udocker.utils.filebind.os.path.exists')
    def test_04_start(self, mock_exists, mock_isfile,
                      mock_realpath, mock_mktmp, mock_setup):
        """Test04 FileBind().start()."""
        container_id = "CONTAINERID"
        mock_realpath.return_value = "/tmp"
        mock_setup.return_value = None
        files_list = ["file1", "dir1", "file2"]
        mock_mktmp.return_value = "tmpDir"
        fbind = FileBind(self.local, container_id)
        fbind.start(files_list)
        self.assertTrue(mock_mktmp.called)

        mock_isfile.side_effect = [True, False, True]
        mock_setup.return_value = None
        fbind = FileBind(self.local, container_id)
        self.assertTrue(mock_isfile.called)
        self.assertTrue(mock_exists.called)
        self.assertIsInstance(fbind.start(files_list), tuple)

    @patch('udocker.utils.filebind.os.path.realpath')
    @patch.object(FileBind, 'set_file')
    def test_05_set_list(self, mock_setfile, mock_realpath):
        """Test05 FileBind().set_list()."""
        container_id = "CONTAINERID"
        flist = ['f1', 'f2']
        mock_realpath.return_value = "/tmp"
        mock_setfile.side_effect = [None, None]
        FileBind(self.local, container_id).set_list(flist)
        self.assertTrue(mock_setfile.called)

    @patch('udocker.utils.filebind.FileUtil.copyto')
    @patch('udocker.utils.filebind.os.symlink')
    @patch('udocker.utils.filebind.os.rename')
    @patch('udocker.utils.filebind.os.path.exists')
    @patch('udocker.utils.filebind.os.path.isfile')
    @patch('udocker.utils.filebind.os.path.realpath')
    def test_06_set_file(self, mock_realpath, mock_isfile,
                         mock_exists, mock_rename, mock_sym,
                         mock_futilcopy):
        """Test06 FileBind().set_file()."""
        hfile = 'host_file'
        cfile = 'cont_file'
        container_id = "CONTAINERID"
        mock_realpath.return_value = "/tmp"
        mock_isfile.return_value = False
        mock_exists.return_value = False
        fbind = FileBind(self.local, container_id)
        fbind.set_file(hfile, cfile)
        self.assertFalse(mock_futilcopy.called)

        mock_realpath.return_value = "/tmp"
        mock_isfile.return_value = True
        mock_exists.return_value = False
        fbind = FileBind(self.local, container_id)
        fbind.set_file(hfile, cfile)
        self.assertTrue(mock_rename.called)
        self.assertTrue(mock_sym.called)

        mock_realpath.return_value = "/tmp"
        mock_isfile.return_value = True
        mock_exists.return_value = True
        fbind = FileBind(self.local, container_id)
        fbind.set_file(hfile, cfile)
        self.assertTrue(mock_futilcopy.called)

    @patch('udocker.utils.filebind.FileUtil.remove')
    @patch('udocker.utils.filebind.FileUtil.copyto')
    @patch('udocker.utils.filebind.os.path.realpath')
    def test_07_add_file(self, mock_realpath, mock_futilcp,
                         mock_futilrm):
        """Test07 FileBind().add_file()."""
        container_id = "CONTAINERID"
        mock_realpath.return_value = "/tmp"
        host_file = "host.file"
        container_file = "#container.file"
        fbind = FileBind(self.local, container_id)
        fbind.host_bind_dir = "/tmp"
        fbind.add_file(host_file, container_file)
        self.assertTrue(mock_futilrm.called)
        self.assertTrue(mock_futilcp.called)

    @patch('udocker.utils.filebind.os.path.realpath')
    def test_08_get_path(self, mock_realpath):
        """Test08 FileBind().get_path()."""
        container_id = 'CONTAINERID'
        cont_file = '/dir1/file'
        mock_realpath.return_value = "/tmp"
        fbind = FileBind(self.local, container_id)
        fbind.host_bind_dir = '/dir0'
        status = fbind.get_path(cont_file)
        self.assertEqual(status, '/dir0/#dir1#file')

    # def test_09_finish(self):
    #     """Test09 FileBind().finish()."""


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""Object classes to store Nix environment"""

import os
import sys
import string

from udocker import is_genstr

def get_pair(envstr):
    """Split env=var into key and val"""
    if not (is_genstr(envstr) and envstr):
        return ("", "")
    if '=' in envstr:
        try:
            (key, val) = envstr.split('=', 1)
        except (ValueError, NameError, AttributeError):
            return ("", "")
        key = key.strip()
        val = val.strip()
        #for quote in ("'", '"'):
        #    if quote == val[0]:
        #        val = val.strip(quote)
        #        break
    else:
        key = envstr.strip()
        val = os.getenv(envstr, "")
    if not key or ' ' in key or key[0] in string.digits:
        return ("", "")
    return (key, val)


class UenvIterator(object):
    """Uenv Iterator class"""

    def __init__(self, uenv):
        """Init Uenv iterator"""
        self._uenv = uenv
        if sys.version_info[0] >= 3:
            self._keys = list(uenv.env.keys())
        else:
            self._keys = uenv.env.keys()
        self._index = 0

    def __next__(self):
        """Python 3 next (key, value) of environment"""
        if self._index < len(self._keys):
            key = self._keys[self._index]
            value = self._uenv.env[key]
            self._index += 1
            return (key, value)
        raise StopIteration

    next = __next__ # Python 2


class Uenv(object):
    """Class to store environment variables"""

    def __init__(self, thisenv=None):
        """Init Uenv instance"""
        self.env = {}
        if thisenv and not self.add(thisenv):
            raise ValueError('invalid environment')

    def __iter__(self):
        """Returns the Uenv iterator"""
        return UenvIterator(self)

    def __len__(self):
        """Returns the number of stored env variables"""
        return len(self.env)

    def append(self, envstr):
        """Add string with key=val to Uenv"""
        (key, val) = get_pair(envstr)
        if key:
            self.env[key] = val
        return self

    def appendif(self, envstr):
        """Add string with key=val to Uenv if key non-existent"""
        (key, val) = get_pair(envstr)
        if key and key not in self.env:
            self.env[key] = val
        return self

    def extend(self, envlist):
        """Extend Uenv stored environment with list or dict of key=val"""
        if isinstance(envlist, (list)):
            for envstr in envlist:
                self.append(envstr)
        elif isinstance(envlist, (dict)):
            for key in envlist.keys():
                self.env[key] = envlist[key]
        return self

    def extendif(self, envlist):
        """Extend Uenv environment with list of key=val if key non-existant"""
        if isinstance(envlist, (list)):
            for envstr in envlist:
                self.appendif(envstr)
        elif isinstance(envlist, (dict)):
            for key in envlist.keys():
                if key not in self.env:
                    self.env[key] = envlist[key]
        return self

    def add(self, thisenv):
        """Extend Uenv environment with str or list with key=val"""
        if isinstance(thisenv, (str)):
            self.append(thisenv)
        elif isinstance(thisenv, (list)):
            self.extend(thisenv)
        return self

    def getenv(self, key):
        """Get variable by name"""
        try:
            return self.env[key]
        except KeyError:
            return ""

    def setenv(self, key, val):
        """Set variable name"""
        return self.append(f'{key}={val}')

    def unsetenv(self, key):
        """Delete variable by name"""
        try:
            del self.env[key]
            return True
        except KeyError:
            return False

    def list(self):
        """Get list with environment variables"""
        env_list = []
        for (key, val) in self.env.items():
            env_list.append(f'{key}={val}')
        return env_list

    def dict(self):
        """Get dict with environment"""
        return self.env

    def keys(self):
        """Get list of environment variable names"""
        return self.env.keys()



# -*- coding: utf-8 -*-
"""Superclass for the execution engines"""

import os
import sys
import re
import json

from udocker import is_genstr
from udocker.msg import Msg
from udocker.config import Config
from udocker.utils.fileutil import FileUtil
from udocker.utils.uenv import Uenv
from udocker.utils.uvolume import Uvolume
from udocker.helper.nixauth import NixAuthentication
from udocker.helper.hostinfo import HostInfo
from udocker.container.structure import ContainerStructure
from udocker.utils.filebind import FileBind
from udocker.utils.mountpoint import MountPoint


class ExecutionEngineCommon(object):
    """Docker container execution engine parent class
    Provides the container execution methods that are common to
    the execution drivers.
    """

    # Metadata defaults
    opt = {}                     # Run options
    opt["nometa"] = False            # Don't load metadata
    opt["nosysdirs"] = False         # Bind host dirs
    opt["dri"] = False               # Directories needed for DRI
    opt["bindhome"] = False          # Bind user home dir
    opt["hostenv"] = False           # Pass host env
    opt["hostauth"] = False          # Use hostauth_list
    opt["containerauth"] = False     # Authentication from container
    opt["novol"] = []                # Volume bindings to ignore
    opt["env"] = Uenv()              # Environment from container
    opt["envfile"] = []              # File with environment variables
    opt["vol"] = []                  # Volumes to mount
    opt["cpuset"] = ""               # Container CPU affinity
    opt["user"] = ""                 # User to run in the container
    opt["cwd"] = ""                  # Default dir in the container
    opt["entryp"] = ""               # Container entrypoint
    opt["cmd"] = []                  # Comand to execute
    opt["hostname"] = ""             # Hostname TBD
    opt["domain"] = ""               # Host domainname TBD
    opt["volfrom"] = []              # Mount vol from container TBD
    opt["portsmap"] = []             # Ports mapped in container
    opt["portsexp"] = []             # Ports exposed by container
    opt["devices"] = []              # Devices passed to container
    opt["nobanner"] = False          # Printing of startup banner

    def __init__(self, localrepo, exec_mode):
        self.localrepo = localrepo            # LocalRepository instance
        self.container_id = ""                # Container id
        self.container_dir = ""               # Container directory
        self.container_root = ""              # ROOT of container filesystem
        self.container_names = []             # Container names
        self.imagerepo = None                 # Imagerepo of container image
        self.hostauth_list = ()               # Authentication files to be used
        self.exec_mode = exec_mode            # ExecutionMode instance
        self.mountp = None                    # MountPoint object
        self.executable = ""                  # Executable proot, runc, etc

    def _has_option(self, search_option, arg=None):
        """Check if executable has a given cli option"""
        return HostInfo().cmd_has_option(self.executable, search_option, arg)

    def _get_portsmap(self, by_container=True):
        """List of TCP/IP ports mapped indexed by container port"""
        indexed_portmap = {}
        for portmap in self.opt["portsmap"]:
            pmap = portmap.split(":")
            try:
                host_port = int(pmap[0])
                cont_port = int(pmap[1])
            except (ValueError, TypeError):
                try:
                    host_port = int(pmap[1])
                    cont_port = int(pmap[2])
                except (ValueError, TypeError):
                    host_port = None
                    cont_port = None
            if host_port:
                if by_container:
                    indexed_portmap[cont_port] = host_port
                else:
                    indexed_portmap[host_port] = cont_port
        return indexed_portmap

    def _check_exposed_ports(self):
        """TCP/UDP ports < 1024 in ExposedPorts JSON metadata
        The exposed ports format is  ExposedPorts:{ "80/tcp":{}, }
        """
        mapped_ports = self._get_portsmap()
        exposes_priv = False
        exposes_port = False
        for port in self.opt["portsexp"]:
            try:
                port_number = int(port.split("/")[0])
                exposes_port = True
            except (ValueError, TypeError):
                pass
            else:
                if port_number < 1024:
                    if port_number in list(mapped_ports.keys()):
                        if mapped_ports[port_number] >= 1024:
                            continue
                    exposes_priv = True
        if exposes_priv and HostInfo.uid != 0:
            Msg().err("Error: this container exposes privileged TCP/IP ports")
            return False
        if exposes_port:
            Msg().out("Warning: this container exposes TCP/IP ports", l=Msg.WAR)
        return True

    def _set_cpu_affinity(self):
        """Set the cpu affinity string for container run command"""
        # find set affinity executable
        if not self.opt["cpuset"]:
            return []
        exec_cmd = []
        for exec_cmd in Config.conf['cpu_affinity_exec_tools']:
            exec_name = \
                FileUtil(exec_cmd[0]).find_exec()
            if exec_name:
                exec_cmd[0] = exec_name
                for (index, arg) in enumerate(exec_cmd):
                    if arg == "%s":
                        exec_cmd[index] = self.opt["cpuset"]
                return exec_cmd
        self.opt["cpuset"] = ""
        return []

    def _create_mountpoint(self, host_path, cont_path, dirs_only=False):
        """Create mountpoint"""
        if dirs_only and not FileUtil(host_path).isdir():
            return True
        if self.mountp.create(host_path, cont_path):
            self.mountp.save(cont_path)
            return True
        return False

    def _check_volumes(self):
        """Check volume paths"""
        for vol in list(self.opt["vol"]):
            (host_path, cont_path) = Uvolume(vol).split()
            if not (host_path and host_path.startswith('/')):
                Msg().err("Error: invalid host volume path:", host_path)
                return False
            if not (cont_path and cont_path != '/' and
                    cont_path.startswith('/')):
                Msg().err("Error: invalid container volume path:", cont_path)
                return False
            if not os.path.exists(host_path):
                if (host_path in Config.conf['dri_list'] or
                        host_path in Config.conf['sysdirs_list']):
                    self.opt["vol"].remove(vol)
                    continue
                Msg().err("Error: invalid host volume path:", host_path)
                return False
            if not self._create_mountpoint(host_path, cont_path):
                Msg().err("Error: creating mountpoint:", host_path, cont_path)
                return False
        return True

    def _get_bindhome(self):
        """Binding of the host $HOME in to the container $HOME"""
        if self.opt["bindhome"]:
            return NixAuthentication().get_home()
        return ""

    def _is_volume(self, path):
        """Is path a host_path in the volumes list"""
        for vol in list(self.opt["vol"]):
            (host_path, cont_path) = Uvolume(vol).split()
            if host_path and host_path == Uvolume().cleanpath(path):
                return cont_path
        return ""

    def _is_mountpoint(self, path):
        """Is path a host_path in the volumes list"""
        for vol in list(self.opt["vol"]):
            (host_path, cont_path) = Uvolume(vol).split()
            if cont_path and cont_path == Uvolume().cleanpath(path):
                return host_path
        return ""

    def _set_volume_bindings(self):
        """Set the volume bindings string for container run command"""
        # predefined volume bindings --sysdirs --hostauth --dri
        if not self.opt["nosysdirs"]:
            self.opt["vol"].extend(Config.conf['sysdirs_list'])
        if self.opt["hostauth"]:
            self.opt["vol"].extend(self.hostauth_list)
        if self.opt["dri"]:
            self.opt["vol"].extend(Config.conf['dri_list'])
        home = self._get_bindhome()
        if home:
            self.opt["vol"].append(home)
        # remove directory bindings specified with --novol
        for novolume in list(self.opt["novol"]):
            found = False
            for vol in list(self.opt["vol"]):
                if novolume in (vol, vol.split(':')[0]):
                    self.opt["vol"].remove(vol)
                    found = True
            if not found:
                Msg().err(f"Warning: --novol {novolume} not in volumes list", l=Msg.WAR)
        return self._check_volumes()

    def _check_paths(self):
        """Make sure we have a reasonable default PATH and CWD"""
        if not self.opt["env"].getenv("PATH"):
            if self.opt["uid"] == "0":
                path = Config.conf['root_path']
            else:
                path = Config.conf['user_path']
            self.opt["env"].append(f"PATH={path}")
        # verify if the working directory is valid and fix it
        if not self.opt["cwd"]:
            self.opt["cwd"] = self.opt["home"]
        cwd_path = FileUtil(self.container_root).cont2host(self.opt["cwd"],
                                                           self.opt["vol"])
        if os.path.isdir(cwd_path):
            return True
        Msg().err("Error: invalid working directory: ", self.opt["cwd"])
        return False

    def _check_executable(self):
        """Check if executable exists and has execute permissions"""
        if self.opt["entryp"] and is_genstr(self.opt["entryp"]):
            self.opt["entryp"] = self.opt["entryp"].strip().split(' ')
        if isinstance(self.opt["entryp"], list):
            if self.opt["cmd"]:                                     # and cmd
                cmd_args = self.opt["cmd"]
                self.opt["cmd"] = self.opt["entryp"]
                self.opt["cmd"].extend(cmd_args)   # cmd=args entryp
            else:
                self.opt["cmd"] = self.opt["entryp"]
        if not self.opt["cmd"]:
            self.opt["cmd"] = Config.conf['cmd']
            Msg().err("Warning: no command assuming:", self.opt["cmd"],
                      l=Msg.WAR)
        exec_name = self.opt["cmd"][0]            # exec pathname without args
        if exec_name.startswith("./") or exec_name.startswith("../"):
            exec_name = self.opt["cwd"] + '/' + exec_name

        path = self.opt["env"].getenv("PATH")
        # DEBUG
        exec_name = FileUtil(exec_name).find_exec(path, self.container_root,
                                                  self.opt["vol"],
                                                  self.opt["cwd"])
        if exec_name:
            return self.container_root + '/' + exec_name
        Msg().err("Error: command not found or has no execute bit set: ",
                  self.opt["cmd"])
        return ""

    def _run_load_metadata(self, container_id):
        """Load container metadata from container JSON payload"""
        # get container metadata unless we are dealing with a simple directory
        # tree in which case we don't have metadata
        if Config.conf['location']:
            container_dir = ""
            container_json = []
        else:
            container_structure = \
                ContainerStructure(self.localrepo, container_id)
            (container_dir, container_json) = \
                container_structure.get_container_attr()
            if not container_dir:
                return(None, None)
            # load metadata from container
            if not self.opt["nometa"]:
                if not self.opt["user"]:
                    self.opt["user"] = \
                        container_structure.get_container_meta(
                            "User", "", container_json)
                if not self.opt["cwd"]:
                    self.opt["cwd"] = \
                        container_structure.get_container_meta(
                            "WorkingDir", "", container_json)
                if not self.opt["hostname"]:
                    self.opt["hostname"] = \
                        container_structure.get_container_meta(
                            "Hostname", "", container_json)
                if not self.opt["domain"]:
                    self.opt["domain"] = \
                        container_structure.get_container_meta(
                            "Domainname", "", container_json)
                if self.opt["entryp"] is False:
                    self.opt["entryp"] = \
                        container_structure.get_container_meta(
                            "Entrypoint", [], container_json)
                    if not self.opt["cmd"]:
                        self.opt["cmd"] = \
                            container_structure.get_container_meta(
                                "Cmd", [], container_json)
                elif not self.opt["entryp"]:
                    self.opt["entryp"] = []
                else:
                    if isinstance(self.opt["entryp"], str):
                        self.opt["entryp"] = \
                            self.opt["entryp"].strip().split(' ')
                self.opt["Volumes"] = \
                    container_structure.get_container_meta(
                        "Volumes", [], container_json)
                self.opt["portsexp"].extend(
                    container_structure.get_container_meta(
                        "ExposedPorts", [], container_json))
                self.opt["env"].extendif(
                    container_structure.get_container_meta(
                        "Env", [], container_json))
        return(container_dir, container_json)

    def _select_auth_files(self):
        """Select authentication files to use /etc/passwd /etc/group"""
        cont_passwd = self.container_root + "/etc/passwd"
        cont_group = self.container_root + "/etc/group"
        bind_passwd = self.container_dir + FileBind.orig_dir + "/#etc#passwd"
        bind_group = self.container_dir + FileBind.orig_dir + "/#etc#group"
        #
        if os.path.islink(cont_passwd) and os.path.isfile(bind_passwd):
            passwd = bind_passwd
        else:
            passwd = cont_passwd
        if os.path.islink(cont_group) and os.path.isfile(bind_group):
            group = bind_group
        else:
            group = cont_group
        if self._is_mountpoint("/etc/passwd"):
            passwd = self._is_mountpoint("/etc/passwd")
        if self._is_mountpoint("/etc/group"):
            group = self._is_mountpoint("/etc/group")
        return (passwd, group)

    def _validate_user_str(self, user):
        """Parse string with uid:gid or username"""
        user_id = {}
        if not is_genstr(user):
            return user_id

        if re.match("^[a-zA-Z_][a-zA-Z0-9_-]*$", user):
            user_id["user"] = user
            return user_id

        match = re.match("^(\\d+)(:(\\d+)){0,1}$", user)
        if match:
            user_id["uid"] = match.group(1)
            if match.group(3):
                user_id["gid"] = match.group(3)

        return user_id

    def _user_from_str(self, user, host_auth=None, container_auth=None):
        """user password entry from host or container passwd
           Returns (valid_user: bool, user_passwd_fields: dict)
        """
        user_id = self._validate_user_str(user)
        if not user_id:
            return (False, user_id)
        search_field = user_id.get("uid", user_id.get("user", ""))
        if self.opt["hostauth"]:
            (self.opt["user"], self.opt["uid"], self.opt["gid"],
             self.opt["gecos"], self.opt["home"], self.opt["shell"]) = \
                host_auth.get_user(search_field)
        else:
            (self.opt["user"], self.opt["uid"], self.opt["gid"],
             self.opt["gecos"], self.opt["home"], self.opt["shell"]) = \
                container_auth.get_user(search_field)
        return (True, user_id)

    def _setup_container_user(self, user):
        """Once we know which username to use inside the container
        we need to check if it exists in the passwd file that will
        be used inside the container. Since we can override the
        usage of the container /etc/passwd and /etc/group with
        files passed from the host, then we must check if this
        username is in the appropriate file so:
        1. check if the passwd will be the one of the host system
           either because we passwd --hostauth or because we did
           --volume=/etc/passwd:/etc/passwd
        2. else we are using the container original /etc/passwd
        In either case the user specified may not exist, in which
        case we copy the passwd file to a new file and we create
        the intended user. The file is then passwd/mapped into
        the container.
        """
        host_auth = NixAuthentication()
        (passwd, group) = self._select_auth_files()
        container_auth = NixAuthentication(passwd, group)
        if not user:
            user = "root"
        (valid_user, user_id) = self._user_from_str(user,
                                                    host_auth, container_auth)
        if not valid_user:
            Msg().err("Error: invalid syntax for user", user)
            return False
        if (self._is_mountpoint("/etc/passwd") or
                self._is_mountpoint("/etc/group")):
            self.opt["hostauth"] = self.opt["containerauth"] = False
            return True
        if self.opt["user"]:
            if self.opt["user"] != "root":
                self.opt["uid"] = str(HostInfo.uid)
                self.opt["gid"] = str(HostInfo.gid)
        else:
            if self.opt["hostauth"] or self.opt["containerauth"]:
                Msg().err("Error: user not found")
                return False
            self.opt["user"] = user_id["user"] if "user" in user_id else user
            self.opt["uid"] = user_id["uid"] if "uid" in user_id else ""
            self.opt["gid"] = user_id["gid"] if "gid" in user_id else ""
        self._create_user(container_auth, host_auth)
        return True

    def _setup_container_user_noroot(self, user):
        """ Setup user for engines without root support.
        Equivalent to _setup_container_user() for engines without root support.
        """
        host_auth = NixAuthentication()
        (passwd, group) = self._select_auth_files()
        container_auth = NixAuthentication(passwd, group)
        if not user:
            user = HostInfo().username()
        (valid_user, user_id) = self._user_from_str(user,
                                                    host_auth, container_auth)
        if not valid_user:
            Msg().err("Error: invalid syntax for user", user)
            return False
        if self.opt["user"] == "root":
            self.opt["user"] = HostInfo().username()
            self.opt["uid"] = str(HostInfo.uid)
            self.opt["gid"] = str(HostInfo.gid)
        if (self._is_mountpoint("/etc/passwd") or
                self._is_mountpoint("/etc/group")):
            self.opt["hostauth"] = self.opt["containerauth"] = False
            return True
        if self.opt["user"]:
            self.opt["uid"] = str(HostInfo.uid)
            self.opt["gid"] = str(HostInfo.gid)
        else:
            if self.opt["hostauth"] or self.opt["containerauth"]:
                Msg().err("Error: user not found on host")
                return False
            self.opt["user"] = user_id["user"] if "user" in user_id else user
            self.opt["uid"] = user_id["uid"] if "uid" in user_id else ""
            self.opt["gid"] = user_id["gid"] if "gid" in user_id else ""
        self._create_user(container_auth, host_auth)
        return True

    def _fill_user(self):
        """Fill in values for user to be used in the account creation.
        Provide default values in case the required fields are empty.
        """
        if not self.opt["uid"]:
            self.opt["uid"] = str(HostInfo.uid)
        if not self.opt["gid"]:
            self.opt["gid"] = str(HostInfo.gid)
        if not self.opt["user"]:
            self.opt["user"] = "udoc" + self.opt["uid"][0:4]
        if self.opt["bindhome"]:
            self.opt["home"] = NixAuthentication().get_home()
        if not self.opt["home"]:
            self.opt["home"] = '/'
        if not self.opt["shell"]:
            self.opt["shell"] = "/bin/sh"
        if not self.opt["gecos"]:
            self.opt["gecos"] = "*UDOCKER*"

    def _create_user(self, container_auth, host_auth):
        """If we need to create a new user then we first
        copy /etc/passwd and /etc/group to new files and them
        we add the user account into these copied files which
        later are binding/mapped/passed to the container. So
        setup this binding as well via hostauth.
        """
        if self.opt["containerauth"]:
            tmp_passwd = container_auth.passwd_file
            tmp_group = container_auth.group_file
            self.opt["hostauth"] = False
        else:
            FileUtil().umask(0o077)
            tmp_passwd = FileUtil("passwd").mktmp()
            tmp_group = FileUtil("group").mktmp()
            if self.opt["hostauth"]:
                FileUtil("/etc/passwd").copyto(tmp_passwd)
                FileUtil("/etc/group").copyto(tmp_group)
            else:
                FileUtil(container_auth.passwd_file).copyto(tmp_passwd)
                FileUtil(container_auth.group_file).copyto(tmp_group)
            FileUtil().umask()
        if not (self.opt["containerauth"] or self.opt["hostauth"]):
            Msg().out("Warning: non-existing user will be created",
                      l=Msg.DBG)
            self._fill_user()
            new_auth = NixAuthentication(tmp_passwd, tmp_group)
            new_auth.add_user(self.opt["user"], 'x',
                              self.opt["uid"], self.opt["gid"],
                              self.opt["gecos"], self.opt["home"],
                              self.opt["shell"])
            (group, dum1, dum2) = host_auth.get_group(self.opt["gid"])
            if not group:
                new_auth.add_group(self.opt["user"], self.opt["gid"])
            for sup_gid in os.getgroups():
                new_auth.add_group('G' + str(sup_gid), str(sup_gid),
                                   [self.opt["user"], ])
        if not self.opt["containerauth"]:
            self.opt["hostauth"] = True
            self.hostauth_list = (tmp_passwd + ":/etc/passwd",
                                  tmp_group + ":/etc/group")
        return True

    def _run_banner(self, cmd, char="*"):
        """Print a container startup banner"""
        if not self.opt["nobanner"]:
            Msg().out("",
                      "\n", char * 78,
                      "\n", char, " " * 74, char,
                      "\n", char,
                      ("STARTING " + self.container_id).center(74, " "), char,
                      "\n", char, " " * 74, char,
                      "\n", char * 78, "\n",
                      "executing:", os.path.basename(cmd), l=Msg.INF)

    def _run_env_cleanup_dict(self):
        """Allow only to pass essential environment variables."""
        environ_copy = os.environ.copy()
        for env_var in environ_copy:
            if env_var in Config.conf['invalid_host_env']:
                del os.environ[env_var]
                continue
            if not self.opt["hostenv"]:
                if env_var not in Config.conf['valid_host_env']:
                    del os.environ[env_var]

    def _run_env_cleanup_list(self):
        """ Allow only to pass essential environment variables.
            Overriding parent ExecutionEngineCommon() class.
        """
        container_env = self.opt["env"].keys()
        for (env_var, value) in list(os.environ.items()):
            if not env_var:
                continue
            if   (env_var in Config.conf['invalid_host_env'] or
                  env_var in container_env):
                continue
            if   ((not self.opt["hostenv"]) and
                  env_var not in Config.conf['valid_host_env']):
                continue
            self.opt["env"].append(f"{env_var}={value}")

    def _run_env_set(self):
        """Environment variables to set"""
        self.opt["env"].appendif("HOME=" + self.opt["home"])
        self.opt["env"].append("USER=" + self.opt["user"])
        self.opt["env"].append("LOGNAME=" + self.opt["user"])
        self.opt["env"].append("USERNAME=" + self.opt["user"])

        if str(self.opt["uid"]) == "0":
            self.opt["env"].append(r"PS1=%s# " % self.container_id[:8])
        else:
            self.opt["env"].append(r"PS1=%s\$ " % self.container_id[:8])

        self.opt["env"].append("SHLVL=0")
        self.opt["env"].append("container_ruser=" + HostInfo().username())
        self.opt["env"].append("container_root=" + self.container_root)
        self.opt["env"].append("container_uuid=" + self.container_id)
        self.opt["env"].append("container_execmode=" +
                               self.exec_mode.get_mode())
        cont_name = self.container_names
        # if Python 3
        if sys.version_info[0] >= 3:
            names = str(cont_name).translate(str.maketrans('', '', " '\"[]"))
        else:
            names = str(cont_name).translate(None, " '\"[]")

        self.opt["env"].append("container_names=" + names)

    def _run_env_cmdoptions(self):
        """Load environment from file --env-file="""
        for envfile in self.opt["envfile"]:
            envdata = FileUtil(envfile).getdata('r')
            for line in envdata.split("\n"):
                self.opt["env"].appendif(line)

    def _run_init(self, container_id):
        """Prepare execution of the container
        To be called by the run() method of the actual ExecutionEngine
        """
        try:
            (container_dir, dummy) = \
                self._run_load_metadata(container_id)
        except (ValueError, TypeError):
            return ""

        if not (container_dir or Config.conf['location']):
            return ""

        self._run_env_cmdoptions()

        if Config.conf['location']:               # using specific root tree
            self.container_root = Config.conf['location']
        else:
            self.container_root = container_dir + "/ROOT"

        # container name(s) from container_id
        self.container_names = self.localrepo.get_container_name(container_id)
        self.container_id = container_id
        self.container_dir = container_dir

        # execution mode
        #self.exec_mode = ExecutionMode(self.localrepo, self.container_id)

        # check if exposing privileged ports
        self._check_exposed_ports()

        # which user to use inside the container and setup its account
        if not self._setup_container_user(self.opt["user"]):
            return ""

        # setup mountpoints
        self.mountp = MountPoint(self.localrepo, self.container_id)

        if not self._set_volume_bindings():
            return ""

        if not self._check_paths():
            return ""

        exec_path = self._check_executable()

        return exec_path

    def _is_same_osenv(self, filename):
        """Check if the host has changed"""
        try:
            saved = json.loads(FileUtil(filename).getdata())
            if (saved["osversion"] == HostInfo().osversion() and
                    saved["oskernel"] == HostInfo().oskernel() and
                    saved["arch"] == HostInfo().arch() and
                    saved["osdistribution"] == \
                        str(HostInfo().osdistribution())):
                return saved
        except (IOError, OSError, AttributeError, ValueError, TypeError,
                IndexError, KeyError):
            pass
        return {}

    def _save_osenv(self, filename, save=None):
        """Save host info for is_same_host()"""
        if save is None:
            save = {}
        try:
            save["osversion"] = HostInfo().osversion()
            save["oskernel"] = HostInfo().oskernel()
            save["arch"] = HostInfo().arch()
            save["osdistribution"] = str(HostInfo().osdistribution())
            if FileUtil(filename).putdata(json.dumps(save)):
                return True
        except (AttributeError, ValueError, TypeError,
                IndexError, KeyError):
            pass
        return False



#!/usr/bin/env python
"""
udocker unit tests: Msg
"""
import sys
from unittest import TestCase, main
from unittest.mock import patch
from io import StringIO
from udocker.msg import Msg

STDOUT = sys.stdout
STDERR = sys.stderr
UDOCKER_TOPDIR = "test_topdir"
BUILTINS = "builtins"


def is_writable_file(obj):
    """Check if obj is a file."""
    try:
        obj.write("")
    except(AttributeError, OSError, IOError):
        return False
    else:
        return True


class MsgTestCase(TestCase):
    """Test Msg() class screen error and info messages."""

    def _verify_descriptors(self, msg):
        """Verify Msg() file descriptors."""
        self.assertTrue(is_writable_file(msg.chlderr))
        self.assertTrue(is_writable_file(msg.chldout))
        self.assertTrue(is_writable_file(msg.chldnul))

    def test_01_init(self):
        """Test01 Msg() constructor."""
        msg = Msg(0)
        self._verify_descriptors(msg)
        self.assertEqual(msg.level, 0)

    def test_02_setlevel(self):
        """Test02 Msg.setlevel()."""
        msg = Msg(5)
        self._verify_descriptors(msg)
        self.assertEqual(msg.level, 5)

        msg = Msg(0)
        msg.setlevel(7)
        self._verify_descriptors(msg)
        self.assertEqual(msg.level, 7)

    @patch('sys.stdout', new_callable=StringIO)
    def test_03_out(self, mock_stdout):
        """Test03 Msg.out()."""
        msg = Msg(Msg.MSG)
        msg.out("111", "222", "333", 444, ('555'))
        self.assertEqual("111 222 333 444 555\n", mock_stdout.getvalue())
        sys.stdout = STDOUT
        sys.stderr = STDERR

    @patch('sys.stderr', new_callable=StringIO)
    def test_04_err(self, mock_stderr):
        """Test04 Msg.err()."""
        msg = Msg(Msg.ERR)
        msg.err("111", "222", "333", 444, ('555'))
        self.assertEqual("111 222 333 444 555\n", mock_stderr.getvalue())
        sys.stdout = STDOUT
        sys.stderr = STDERR


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""LocalFile API integration for Docker and OCI file manipulation"""

import os

from udocker.msg import Msg
from udocker.utils.fileutil import FileUtil

from udocker.docker import DockerLocalFileAPI
from udocker.oci import OciLocalFileAPI
from udocker.commonlocalfile import CommonLocalFileApi


class LocalFileAPI(CommonLocalFileApi):
    """Generic manipulation of containers and/or image files"""

    def __init__(self, localrepo):
        CommonLocalFileApi.__init__(self, localrepo)

    def load(self, imagefile, imagerepo=None):
        """Generic load of image tags to a file"""
        if not os.path.exists(imagefile) and imagefile != '-':
            Msg().err("Error: image file does not exist:", imagefile)
            return False
        tmp_imagedir = FileUtil("load").mktmp()
        try:
            os.makedirs(tmp_imagedir)
        except (IOError, OSError):
            return False
        if not self._untar_saved_container(imagefile, tmp_imagedir):
            Msg().err("Error: failed to extract container:", imagefile)
            FileUtil(tmp_imagedir).remove(recursive=True)
            return False
        imagetype = self._get_imagedir_type(tmp_imagedir)
        if imagetype == "Docker":
            repositories = DockerLocalFileAPI(
                self.localrepo).load(tmp_imagedir, imagerepo)
        elif imagetype == "OCI":
            repositories = OciLocalFileAPI(
                self.localrepo).load(tmp_imagedir, imagerepo)
        else:
            repositories = []
        FileUtil(tmp_imagedir).remove(recursive=True)
        return repositories

    def save(self, imagetag_list, imagefile):
        """Generic save of image tags to a file"""
        return DockerLocalFileAPI(self.localrepo).save(
            imagetag_list, imagefile)



# -*- coding: utf-8 -*-
"""
========
udocker
========
Wrapper to execute basic docker containers without using docker.
This tool is a last resort for the execution of docker containers
where docker is unavailable. It only provides a limited set of
functionalities.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import sys
__author__ = "udocker@lip.pt"
__copyright__ = "Copyright 2016 - 2022, LIP"
__credits__ = ["PRoot http://proot.me",
               "runC https://runc.io",
               "crun https://github.com/containers/crun",
               "Fakechroot https://github.com/dex4er/fakechroot",
               "Singularity http://singularity.lbl.gov"
              ]
__license__ = "Licensed under the Apache License, Version 2.0"
__version__ = "1.3.4"
__date__ = "2022"

def is_genstr(objstr):
    """test if objstr is string or unicode both in py2 nd py3
    unicode type has been removed in py3
    :param objstr: object to test if string or unicode
    :return (bool) is_gstr if it is string or unicode or not
    """
    is_gstr = False
    if sys.version_info[0] >= 3:
        is_gstr = isinstance(objstr, str)
    else:
        is_gstr = isinstance(objstr, (str, unicode))
    return is_gstr



#!/usr/bin/env python
"""
udocker unit tests: Unshare
"""

from unittest import TestCase, main
from unittest.mock import patch, MagicMock
from udocker.helper.unshare import Unshare


class UnshareTestCase(TestCase):
    """Test Unshare()."""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    @patch('udocker.helper.unshare.Msg.err')
    @patch('udocker.helper.unshare.ctypes.CDLL')
    def test_01_unshare(self, mock_cdll, mock_msg):
        """Test01 Unshare().unshare"""
        mock_msg.level = 0
        status = Unshare().unshare(False)
        self.assertTrue(mock_cdll.return_value.unshare.called)
        self.assertTrue(status)

        mock_cdll.return_value.unshare.return_value = -1
        status = Unshare().unshare(True)
        self.assertTrue(mock_cdll.return_value.unshare.called)
        self.assertFalse(status)

    @patch('udocker.helper.unshare.os._exit')
    @patch.object(Unshare, 'unshare')
    @patch('udocker.helper.unshare.os.setgroups')
    @patch('udocker.helper.unshare.os.setuid')
    @patch('udocker.helper.unshare.os.setgid')
    @patch('udocker.helper.unshare.Msg.err')
    @patch('udocker.helper.unshare.os.waitpid')
    @patch('udocker.helper.unshare.NixAuthentication.user_in_subgid')
    @patch('udocker.helper.unshare.subprocess.call')
    @patch('udocker.helper.unshare.NixAuthentication.user_in_subuid')
    @patch('udocker.helper.unshare.HostInfo')
    @patch('udocker.helper.unshare.os.read')
    @patch('udocker.helper.unshare.os.close')
    @patch('udocker.helper.unshare.os.fork')
    @patch('udocker.helper.unshare.os.pipe')
    def test_02_namespace_exec(self, mock_pipe, mock_fork, mock_close,
                               mock_read, mock_hinfo, mock_usubuid,
                               mock_call, mock_usubgid, mock_wait, mock_msg,
                               mock_setgid, moc_setuid, mock_setgrp,
                               mock_ush, mock_exit):
        """Test02 Unshare().namespace_exec"""

        # cpid exists waitpid=0
        mock_msg.level = 0
        mock_pipe.side_effect = [('rfid1', 'wfid1'), ('rfid2', 'wfid2')]
        mock_fork.return_value = 1234
        mock_close.side_effect = [None, None]
        mock_read.return_value = None
        mock_hinfo.return_value.username.return_value = 'user'
        mock_hinfo.uid = 1000
        mock_usubuid.return_value = [(1000, 1000)]
        mock_usubgid.return_value = [(1000, 1000)]
        mock_call.side_effect = [None, None]
        mock_wait.return_value = (123, 0)
        status = Unshare().namespace_exec('1')
        self.assertEqual(mock_pipe.call_count, 2)
        self.assertTrue(mock_fork.called)
        self.assertEqual(mock_close.call_count, 2)
        self.assertTrue(mock_hinfo.return_value.username.called)
        self.assertTrue(mock_usubuid.called)
        self.assertTrue(mock_usubgid.called)
        self.assertEqual(mock_call.call_count, 2)
        self.assertFalse(mock_msg.called)
        self.assertTrue(status)

        # cpid not exists
        mock_method = MagicMock(name='method')
        mock_pipe.side_effect = [('rfid1', 'wfid1'), ('rfid2', 'wfid2')]
        mock_fork.return_value = None
        mock_ush.return_value = None
        mock_close.side_effect = [None, None]
        mock_read.return_value = None
        mock_setgid.return_value = None
        moc_setuid.return_value = None
        mock_setgrp.return_value = None
        mock_exit.return_value = 1
        status = Unshare().namespace_exec(mock_method)
        self.assertTrue(status)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""Singularity execution engine"""

import sys
import os
import subprocess

from udocker.engine.base import ExecutionEngineCommon
from udocker.msg import Msg
from udocker.config import Config
from udocker.utils.fileutil import FileUtil
from udocker.utils.uvolume import Uvolume
from udocker.helper.nixauth import NixAuthentication
from udocker.utils.filebind import FileBind
from udocker.helper.unique import Unique
from udocker.helper.hostinfo import HostInfo

class SingularityEngine(ExecutionEngineCommon):
    """Docker container execution engine using singularity
    Provides a namespaces based user space container.
    Inherits from ContainerEngine class
    """

    def __init__(self, localrepo, exec_mode):
        super(SingularityEngine, self).__init__(localrepo, exec_mode)
        self.executable = None                   # singularity
        self.execution_id = None

    def select_singularity(self):
        """Set singularity executable and related variables"""
        self.executable = Config.conf['use_singularity_executable']
        if self.executable != "UDOCKER" and not self.executable:
            self.executable = FileUtil("singularity").find_exec()

        if self.executable == "UDOCKER" or not self.executable:
            self.executable = ""
            arch = HostInfo().arch()
            image_list = []
            if arch == "amd64":
                image_list = ["singularity-x86_64", "singularity"]
            elif arch == "i386":
                image_list = ["singularity-x86", "singularity"]
            elif arch == "arm64":
                image_list = ["singularity-arm64", "singularity"]
            elif arch == "arm":
                image_list = ["singularity-arm", "singularity"]

            f_util = FileUtil(self.localrepo.bindir)
            self.executable = f_util.find_file_in_dir(image_list)

        if not os.path.exists(self.executable):
            Msg().err("Error: singularity executable not found")
            sys.exit(1)

    def _get_volume_bindings(self):
        """Get the volume bindings string for singularity exec"""
        vol_list = []
        home_dir = NixAuthentication().get_home()
        home_is_binded = False
        tmp_is_binded = False
        vartmp_is_binded = False
        for vol in self.opt["vol"]:
            (host_path, cont_path) = Uvolume(vol).split()
            if os.path.isdir(host_path):
                if host_path == home_dir and cont_path in ("", host_path):
                    home_is_binded = True
                elif host_path == "/tmp" and cont_path in ("", "/tmp"):
                    tmp_is_binded = True
                elif host_path == "/var/tmp" and cont_path in ("", "/var/tmp"):
                    vartmp_is_binded = True
            vol_list.extend(["-B", f"{host_path}:{cont_path}", ])
        if not home_is_binded:
            vol_list.extend(["--home", f"{self.container_root}/root:/root", ])
        if not tmp_is_binded:
            vol_list.extend(["-B", f"{self.container_root}/tmp:/tmp", ])
        if not vartmp_is_binded:
            vol_list.extend(["-B", f"{self.container_root}/var/tmp:/var/tmp", ])
        return vol_list

    def _singularity_env_get(self):
        """Build environment string with user specified environment in
        the form SINGULARITYENV_var=value
        """
        singularityenv = {}
        for (key, val) in self.opt["env"]:
            singularityenv[f'SINGULARITYENV_{key}'] = val
        return singularityenv

    def _make_container_directories(self):
        """Create directories expected by Singularity"""
        FileUtil(self.container_root + "/var/tmp").mkdir()
        FileUtil(self.container_root + "/tmp").mkdir()
        FileUtil(self.container_root + "/proc").mkdir()
        FileUtil(self.container_root + "/dev").mkdir()
        FileUtil(self.container_root + "/sys").mkdir()
        FileUtil(self.container_root + "/root").mkdir()

    def _run_invalid_options(self):
        """check -p --publish -P --publish-all --net-coop"""
        if self.opt["portsmap"]:
            Msg().out("Warning: this execution mode does not support "
                      "-p --publish", l=Msg.WAR)
        if self.opt["netcoop"]:
            Msg().out("Warning: this execution mode does not support "
                      "-P --netcoop --publish-all", l=Msg.WAR)

    def _run_as_root(self):
        """Set configure running as normal user or as root via --fakeroot
        """
        username = HostInfo().username()
        if "user" in self.opt:
            if self.opt["user"] == username:
                return False

            if self.opt["user"] != "root" and self.opt["uid"] != '0':
                Msg().out("Warning: running as another user not supported")
                return False

            if self._has_option("--fakeroot", "exec"):
                if (NixAuthentication().user_in_subuid(username) and
                        NixAuthentication().user_in_subgid(username)):
                    Config.conf['singularity_options'].extend(["--fakeroot", ])
                    return True

        self.opt["user"] = username
        return False

    def run(self, container_id):
        """Execute a Docker container using singularity.
        This is the main method invoked to run a container with singularity.
          * argument: container_id or name
          * options:  many via self.opt see the help
        """

        if os.path.isdir(
                FileBind(self.localrepo, container_id).container_orig_dir):
            FileBind(self.localrepo, container_id).restore() # legacy 1.1.3

        Config.conf['sysdirs_list'] = (
            # "/dev", "/proc", "/sys",
            "/etc/resolv.conf", "/etc/host.conf",
            "/lib/modules",
        )

        # setup execution
        exec_path = self._run_init(container_id)
        if not exec_path:
            return 2
        self.opt["cmd"][0] = exec_path.replace(self.container_root + "/", "")

        self._run_invalid_options()

        self._make_container_directories()

        self.select_singularity()

        self._run_as_root()

        # set environment variables
        self._run_env_set()

        if Msg.level >= Msg.DBG:
            singularity_debug = ["--debug", "-v", ]
        elif self._has_option("--silent"):
            singularity_debug = ["--silent", ]
        elif self._has_option("--quiet"):
            singularity_debug = ["--quiet", ]
        else:
            singularity_debug = []

        if self.executable.startswith(self.localrepo.bindir):
            Config.conf['singularity_options'].extend(["-u", ])

        #if FileUtil("nvidia-smi").find_exec():
        #    Config.conf['singularity_options'].extend(["--nv", ])

        singularity_vol_list = self._get_volume_bindings()

        # build the actual command
        self.execution_id = Unique().uuid(self.container_id)
        cmd_l = self._set_cpu_affinity()
        cmd_l.append(self.executable)
        cmd_l.extend(singularity_debug)
        cmd_l.append("exec")
        cmd_l.extend(Config.conf['singularity_options'])
        if self.opt["cwd"]:
            cmd_l.extend(["--pwd", self.opt["cwd"], ])
        cmd_l.extend(singularity_vol_list)
        cmd_l.append(self.container_root)
        cmd_l.extend(self.opt["cmd"])
        Msg().out("CMD =", cmd_l, l=Msg.VER)

        # if not --hostenv clean the environment
        self._run_env_cleanup_dict()

        # execute
        self._run_banner(self.opt["cmd"][0], '/')
        status = subprocess.call(cmd_l, shell=False, close_fds=True, \
            env=os.environ.update(self._singularity_env_get()))
        return status



#!/usr/bin/env python
"""
udocker unit tests: MountPoint
"""

from unittest import TestCase, main
from unittest.mock import Mock, patch
from udocker.utils.mountpoint import MountPoint
from udocker.config import Config


class MountPointTestCase(TestCase):
    """Test MountPoint()."""

    def setUp(self):
        Config().getconf()
        self.bind_dir = "/.bind_host_files"
        self.orig_dir = "/.bind_orig_files"
        str_local = 'udocker.container.localrepo.LocalRepository'
        self.lrepo = patch(str_local)
        self.local = self.lrepo.start()
        self.mock_lrepo = Mock()
        self.local.return_value = self.mock_lrepo

    def tearDown(self):
        self.lrepo.stop()

    @patch.object(MountPoint, 'setup')
    @patch('udocker.utils.mountpoint.os.path.realpath')
    def test_01_init(self, mock_realpath, mock_setup):
        """Test01 MountPoint() constructor"""
        container_id = "CONTAINERID"
        mock_realpath.return_value = "/tmp"
        mock_setup.return_value = True
        mpoint = MountPoint(self.local, container_id)
        self.assertEqual(mpoint.localrepo, self.local)
        self.assertEqual(mpoint.container_id, container_id)
        self.assertTrue(mock_realpath.called)
        self.assertTrue(mpoint.container_root, mpoint.container_dir + "/ROOT")
        self.assertTrue(mock_setup.called)

    @patch('udocker.utils.mountpoint.Msg.err')
    @patch('udocker.utils.mountpoint.FileUtil.mkdir')
    @patch('udocker.utils.mountpoint.os.path.isdir')
    @patch('udocker.utils.mountpoint.os.path.realpath')
    def test_02_setup(self, mock_realpath, mock_isdir, mock_mkdir, mock_msg):
        """Test02 MountPoint().setup()."""
        container_id = "CONTAINERID"
        mock_realpath.return_value = "/tmp"
        mock_isdir.return_value = True
        mpoint = MountPoint(self.local, container_id)
        status = mpoint.setup()
        self.assertTrue(mock_isdir.called)
        self.assertTrue(status)

        mock_isdir.return_value = False
        mock_mkdir.return_value = True
        mpoint = MountPoint(self.local, container_id)
        status = mpoint.setup()
        self.assertTrue(mock_mkdir.called)
        self.assertTrue(status)

        mock_isdir.return_value = False
        mock_mkdir.return_value = False
        mpoint = MountPoint(self.local, container_id)
        status = mpoint.setup()
        self.assertTrue(mock_msg.called)
        self.assertFalse(status)

    @patch('udocker.utils.mountpoint.FileUtil.getvalid_path')
    @patch.object(MountPoint, 'setup')
    @patch('udocker.utils.mountpoint.os.path.realpath')
    def test_03_add(self, mock_realpath, mock_setup, mock_valid):
        """Test03 MountPoint().add()."""
        container_id = "CONTAINERID"
        cpath = 'cont_path'
        mock_realpath.return_value = "/tmp"
        mock_setup.return_value = True
        mock_valid.return_value = "/ROOT/dir1"
        mpoint = MountPoint(self.local, container_id)
        mpoint.add(cpath)
        self.assertEqual(mpoint.mountpoints[cpath], "/ROOT/dir1")

    @patch('udocker.utils.mountpoint.os.path.normpath')
    @patch.object(MountPoint, 'setup')
    @patch('udocker.utils.mountpoint.os.path.realpath')
    def test_04_delete(self, mock_realpath, mock_setup, mock_norm):
        """Test04 MountPoint().delete()."""
        container_id = "CONTAINERID"
        cpath = 'cont_path'
        mock_realpath.return_value = "/tmp"
        mock_setup.return_value = True
        mpoint = MountPoint(self.local, container_id)
        mpoint.mountpoints = {'some_path': '/cont/bin'}
        status = mpoint.delete(cpath)
        self.assertFalse(status)

        mock_realpath.return_value = "/tmp"
        mock_setup.return_value = True
        mock_norm.side_effect = ['/controot/cont_path', '/controot/cont_path']
        mpoint = MountPoint(self.local, container_id)
        mpoint.mountpoints = {'cont_path': '/cont/bin'}
        status = mpoint.delete(cpath)
        self.assertEqual(mpoint.mountpoints, {})
        self.assertTrue(status)

    @patch('udocker.utils.mountpoint.os.path.normpath')
    @patch.object(MountPoint, 'setup')
    @patch('udocker.utils.mountpoint.os.path.realpath')
    def test_05_delete_all(self, mock_realpath, mock_setup, mock_norm):
        """Test05 MountPoint().delete_all()."""
        container_id = "CONTAINERID"
        mock_realpath.return_value = "/tmp"
        mock_setup.return_value = True
        mock_norm.side_effect = ['/controot/cont_path', '/controot/cont_path']
        mpoint = MountPoint(self.local, container_id)
        mpoint.mountpoints = {'cont_path': '/cont/bin'}
        mpoint.delete_all()
        self.assertEqual(mpoint.mountpoints, {})

    @patch.object(MountPoint, 'delete')
    @patch.object(MountPoint, 'add')
    @patch('udocker.utils.mountpoint.Msg.err')
    @patch('udocker.utils.mountpoint.os.path.isdir')
    @patch('udocker.utils.mountpoint.os.path.islink')
    @patch('udocker.utils.mountpoint.FileUtil.putdata')
    @patch('udocker.utils.mountpoint.FileUtil.mkdir')
    @patch('udocker.utils.mountpoint.os.path.isfile')
    @patch('udocker.utils.mountpoint.os.path.exists')
    @patch.object(MountPoint, 'setup')
    @patch('udocker.utils.mountpoint.os.path.realpath')
    def test_06_create(self, mock_realpath, mock_setup, mock_exists,
                       mock_isfile, mock_mkdir, mock_putdata,
                       mock_islink, mock_isdir, mock_msg,
                       mock_add, mock_del):
        """Test06 MountPoint().create()."""
        container_id = "CONTAINERID"
        mock_realpath.return_value = "/tmp"
        mock_setup.return_value = True
        mock_exists.return_value = False
        hpath = '/bin'
        cpath = '/ROOT/bin'
        mock_add.return_value = None
        mock_del.return_value = True
        mock_isfile.side_effect = [True, True]
        mock_mkdir.return_value = False
        mock_putdata.return_value = None
        mock_islink.return_value = True
        mock_isdir.return_value = False
        mpoint = MountPoint(self.local, container_id)
        status = mpoint.create(hpath, cpath)
        self.assertTrue(mock_add.called)
        self.assertTrue(mock_isfile.called)
        self.assertTrue(mock_mkdir.called)
        self.assertTrue(mock_putdata.called)
        self.assertFalse(mock_del.called)
        self.assertTrue(status)

        mock_exists.return_value = False
        mock_add.return_value = None
        mock_del.return_value = True
        mock_isfile.side_effect = [False, False]
        mock_mkdir.return_value = True
        mock_putdata.return_value = None
        mock_islink.return_value = True
        mock_isdir.return_value = True
        mpoint = MountPoint(self.local, container_id)
        status = mpoint.create(hpath, cpath)
        self.assertTrue(mock_mkdir.called)
        self.assertFalse(mock_msg.called)
        self.assertTrue(status)

        mock_exists.return_value = False
        mock_add.return_value = None
        mock_del.return_value = True
        mock_isfile.side_effect = [False, False]
        mock_mkdir.return_value = False
        mock_putdata.return_value = None
        mock_islink.return_value = True
        mock_isdir.return_value = True
        mpoint = MountPoint(self.local, container_id)
        status = mpoint.create(hpath, cpath)
        self.assertTrue(mock_mkdir.called)
        self.assertTrue(mock_msg.called)
        self.assertTrue(mock_del.called)
        self.assertFalse(status)

    @patch('udocker.utils.mountpoint.os.symlink')
    @patch('udocker.utils.mountpoint.os.path.exists')
    @patch.object(MountPoint, 'setup')
    @patch('udocker.utils.mountpoint.os.path.realpath')
    def test_07_save(self, mock_realpath, mock_setup, mock_exists,
                     mock_syml):
        """Test07 MountPoint().save()."""
        container_id = "CONTAINERID"
        cpath = 'cont_path'
        mock_realpath.return_value = "/tmp"
        mock_setup.return_value = True
        mpoint = MountPoint(self.local, container_id)
        mpoint.mountpoints = {'some_path': '/cont/bin'}
        status = mpoint.save(cpath)
        self.assertTrue(status)

        mock_realpath.return_value = "/tmp"
        mock_setup.return_value = True
        mock_exists.return_value = False
        mock_syml.return_value = None
        mpoint = MountPoint(self.local, container_id)
        mpoint.mountpoints = {'cont_path': '/cont/bin'}
        status = mpoint.save(cpath)
        self.assertTrue(status)
        self.assertTrue(mock_exists.called)
        self.assertTrue(mock_syml.called)

    @patch.object(MountPoint, 'save')
    @patch.object(MountPoint, 'setup')
    @patch('udocker.utils.mountpoint.os.path.realpath')
    def test_08_save_all(self, mock_realpath, mock_setup, mock_save):
        """Test08 MountPoint().save_all()."""
        container_id = "CONTAINERID"
        mock_realpath.return_value = "/tmp"
        mock_setup.return_value = True
        mock_save.return_value = None
        mpoint = MountPoint(self.local, container_id)
        mpoint.mountpoints = {'cont_path': '/cont/bin'}
        mpoint.save_all()
        self.assertTrue(mock_save.called)

    @patch('udocker.utils.mountpoint.os.readlink')
    @patch('udocker.utils.mountpoint.os.listdir')
    @patch.object(MountPoint, 'setup')
    @patch('udocker.utils.mountpoint.os.path.realpath')
    def test_09_load_all(self, mock_realpath, mock_setup,
                         mock_ldir, mock_readl):
        """Test09 MountPoint().load_all()."""
        container_id = "CONTAINERID"
        result = {'/dir1': '/dir1', '/dir2': '/dir2'}
        mock_realpath.return_value = "/tmp"
        mock_setup.return_value = True
        mock_ldir.return_value = ['/dir1', '/dir2']
        mock_readl.side_effect = ['/dir1', '/dir2']
        mpoint = MountPoint(self.local, container_id)
        mpoint.load_all()
        self.assertEqual(mpoint.mountpoints, result)

    @patch('udocker.utils.mountpoint.FileUtil.remove')
    @patch.object(MountPoint, 'delete_all')
    @patch.object(MountPoint, 'load_all')
    @patch.object(MountPoint, 'save_all')
    @patch.object(MountPoint, 'setup')
    @patch('udocker.utils.mountpoint.os.path.realpath')
    def test_10_restore(self, mock_realpath, mock_setup, mock_save,
                        mock_load, mock_del, mock_remove):
        """Test10 MountPoint().restore()."""
        container_id = "CONTAINERID"
        mock_realpath.return_value = "/tmp"
        mock_setup.return_value = True
        mock_save.return_value = None
        mock_load.return_value = None
        mock_del.return_value = None
        mock_remove.return_value = None
        mpoint = MountPoint(self.local, container_id)
        mpoint.restore()
        self.assertTrue(mock_save.called)
        self.assertTrue(mock_load.called)
        self.assertTrue(mock_del.called)
        self.assertTrue(mock_remove.called)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""Messages, output, errors, MSG levels"""
import sys


class Msg(object):
    """Write messages to stdout and stderr. Allows to filter the
    messages to be displayed through a verbose level, also allows
    to control if child process that produce output through a
    file descriptor should be redirected to /dev/null
    """

    NIL = -1
    ERR = 0
    MSG = 1
    WAR = 2
    INF = 3
    VER = 4
    DBG = 5
    DEF = INF
    level = DEF
    previous = DEF
    nullfp = None
    chlderr = sys.stderr
    chldout = sys.stdout
    chldnul = sys.stderr

    def __init__(self, new_level=None):
        """
        Initialize Msg level and /dev/null file pointers to be
        used in subprocess calls to obfuscate output and errors
        """
        if new_level is not None:
            Msg.level = new_level
        try:
            if Msg.nullfp is None:
                Msg.nullfp = open('/dev/null', 'w', encoding='utf-8')
        except (IOError, OSError):
            Msg.chlderr = sys.stderr
            Msg.chldout = sys.stdout
            Msg.chldnul = sys.stderr
        else:
            Msg.chlderr = Msg.nullfp
            Msg.chldout = Msg.nullfp
            Msg.chldnul = Msg.nullfp

    def setlevel(self, new_level=None):
        """Define debug level"""
        if new_level is None:
            new_level = Msg.previous
        else:
            Msg.previous = Msg.level
        Msg.level = new_level
        if Msg.level >= Msg.DBG:
            Msg.chlderr = sys.stderr
            Msg.chldout = sys.stdout
        else:
            Msg.chlderr = Msg.nullfp
            Msg.chldout = Msg.nullfp
        return Msg.previous

    def out(self, *args, **kwargs):
        """Write text to stdout respecting verbose level"""
        level = kwargs.get("l", Msg.MSG)
        if level <= Msg.level:
            sys.stdout.write(' '.join([str(x) for x in args]) + '\n')

    def err(self, *args, **kwargs):
        """Write text to stderr respecting verbose level"""
        level = kwargs.get("l", Msg.ERR)
        if level <= Msg.level:
            sys.stderr.write(' '.join([str(x) for x in args]) + '\n')



#!/usr/bin/env python

# -*- coding: utf-8 -*-
"""
udocker unit tests: CmdParser
"""

from unittest import TestCase, main
from udocker.cmdparser import CmdParser


class CmdParserTestCase(TestCase):
    """Test CmdParserTestCase() command line interface."""

    def test_01__init(self):
        """Test01 CmdParser() Constructor."""
        cmdp = CmdParser()
        self.assertEqual(cmdp._argv, "")
        self.assertIsInstance(cmdp._argv_split, dict)
        self.assertIsInstance(cmdp._argv_consumed_options, dict)
        self.assertIsInstance(cmdp._argv_consumed_params, dict)
        self.assertEqual(cmdp._argv_split['CMD'], "")
        self.assertEqual(cmdp._argv_split['GEN_OPT'], [])
        self.assertEqual(cmdp._argv_split['CMD_OPT'], [])
        self.assertEqual(cmdp._argv_consumed_options['GEN_OPT'], [])
        self.assertEqual(cmdp._argv_consumed_options['CMD_OPT'], [])
        self.assertEqual(cmdp._argv_consumed_params['GEN_OPT'], [])
        self.assertEqual(cmdp._argv_consumed_params['CMD_OPT'], [])

    def test_02_parse(self):
        """Test02 CmdParser().parse()."""
        cmdp = CmdParser()
        argv = ["udocker", "run", "--bindhome", "--hostauth", "--hostenv",
                "-v", "/sys", "-v", "/proc", "-v", "/var/run", "-v", "/dev",
                "--user=jorge", "--dri", "myfed", "firefox"]
        status = cmdp.parse(argv)
        self.assertTrue(status)

        argv = ""
        status = cmdp.parse(argv)
        self.assertFalse(status)

    def test_03_missing_options(self):
        """Test03 CmdParser().missing_options()."""
        cmdp = CmdParser()
        argv = ["udocker", "run", "--bindhome", "--hostauth", "--hostenv",
                "-v", "/sys", "-v", "/proc", "-v", "/var/run", "-v", "/dev",
                "--user=jorge", "--dri", "myfed", "firefox"]
        cmdp.parse(argv)
        out = cmdp.missing_options()
        self.assertIsInstance(out, list)

    def test_04_get(self):
        """Test04 CmdParser().get()."""
        argv = ["udocker", "--debug", "run", "--bindhome", "--hostauth",
                "--hostenv", "-v", "/sys", "-v", "/proc", "-v",
                "/var/run", "-v", "/dev", "--user=jorge", "--dri",
                "myfed", "firefox"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        out = cmdp.get("xyz")
        self.assertIsNone(out)

        argv = ["udocker", "--debug", "run", "--bindhome",
                "-v", "/sys", "--user=jorge", "--dri", "myfed", "firefox"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        out = cmdp.get("--debug", "GEN_OPT", True)
        self.assertTrue(out)
        out = cmdp.get("", "CMD", True)
        self.assertTrue(out, "run")
        out = cmdp.get("--bindhome", "CMD_OPT", True)
        self.assertEqual(out, [True])

        argv = ["udocker", "export", "-o", "cont.tar", "12345"]
        cmdp = CmdParser()
        cmdp.parse(argv)
        out = cmdp.get("", "CMD", True)
        cmdp.get("-o")
        tarfile = cmdp.get("P1")
        container_id = cmdp.get("P2")
        self.assertTrue(out, "export")
        self.assertTrue(tarfile, "cont.tar")
        self.assertTrue(container_id, "12345")

    # def test_05_declare_options(self):
    #     """Test05 CmdParser().declare_options()."""

    # def test_06__get_option(self):
    #     """Test06 CmdParser()._get_option()."""

    # def test_07__get_param(self):
    #     """Test07 CmdParser()._get_param()."""

if __name__ == '__main__':
    main()
